from tablate.classes.options.html.style.utilities.style_types import ElementTypeKey, TablateInstanceKey

base_class = "tablate"

styles_key = "styles"

tablate_instance_key: TablateInstanceKey = "tablate_instance"
element_type_key: ElementTypeKey = "element_type"

style_defaults = {

}


from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.initializers.mappers.element.base.base_column_mapper import base_column_mapper
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_text_mapper import base_text_mapper
from tablate.library.initializers.mappers.element.html.html_column_mapper import html_column_mapper
from tablate.library.initializers.mappers.element.html.html_frame_mapper import html_frame_mapper
from tablate.library.initializers.mappers.element.html.html_text_mapper import html_text_mapper
from tablate.type.defaults import html_px_multiplier_default
from tablate.type.primitives import HtmlPxMultiplier
from tablate.type.type_style import FrameStyles, TextStyles, ColumnStyles, HtmlFrameStyles, HtmlColumnStyles, HtmlTextStyles
from tablate.type.type_global import Globals
from tablate.type.type_store import BaseFrameStore, FrameStore, HtmlFrameStore


def style_mapper(base_input: Optional[BaseFrameStore],
                 html_input: Optional[HtmlFrameStore],
                 base_defaults: Optional[BaseFrameStore],
                 html_defaults: Optional[HtmlFrameStore],
                 html_px_multiplier: HtmlPxMultiplier,
                 global_options: Globals) -> FrameStore:
    ####################################################################################################################
    ####################################################################################################################
    ####################################################################################################################

    if html_px_multiplier is None:
        html_px_multiplier = html_px_multiplier_default

    ####################################################################################################################
    ####################################################################################################################
    ####################################################################################################################

    frame_styles_default = base_frame_mapper(frame_input=global_options.console.frame_styles,
                                             frame_defaults=FrameStyles())
    column_styles_default = base_column_mapper(columns_input=global_options.console.column_styles,
                                               column_defaults=ColumnStyles())
    text_styles_default = base_text_mapper(text_input=global_options.console.text_styles,
                                           text_defaults=TextStyles())

    frame_styles_default = set_attr_resolver(instance=base_defaults,
                                             attr="frame_styles",
                                             default=frame_styles_default)
    column_styles_default = set_attr_resolver(instance=base_defaults,
                                              attr="column_styles",
                                              default=column_styles_default)
    text_styles_default = set_attr_resolver(instance=base_defaults,
                                            attr="text_styles",
                                            default=text_styles_default)

    ####################################################################################################################

    frame_styles = set_attr_resolver(instance=base_input,
                                     attr="frame_styles",
                                     default=frame_styles_default)
    column_styles = set_attr_resolver(instance=base_input,
                                      attr="column_styles",
                                      default=column_styles_default)
    text_styles = set_attr_resolver(instance=base_input,
                                    attr="text_styles",
                                    default=text_styles_default)

    frame_styles = base_frame_mapper(frame_input=frame_styles,
                                     frame_defaults=frame_styles_default)
    column_styles = base_column_mapper(columns_input=column_styles,
                                       column_defaults=column_styles_default)
    text_styles = base_text_mapper(text_input=text_styles,
                                   text_defaults=text_styles_default)

    ####################################################################################################################
    ####################################################################################################################
    ####################################################################################################################

    html_frame_styles_default = html_frame_mapper(html_frame_input=global_options.html.html_frame_styles,
                                                  html_frame_defaults=HtmlFrameStyles(),
                                                  html_px_multiplier=html_px_multiplier)
    html_column_styles_default = html_column_mapper(html_columns_input=global_options.html.html_column_styles,
                                                    html_column_defaults=HtmlColumnStyles())
    html_text_styles_default = html_text_mapper(html_text_input=global_options.html.html_text_styles,
                                                html_text_defaults=HtmlTextStyles(),
                                                html_px_multiplier=html_px_multiplier)

    html_frame_styles_default = set_attr_resolver(instance=html_defaults,
                                                  attr="html_frame_styles",
                                                  default=html_frame_styles_default)
    html_column_styles_default = set_attr_resolver(instance=html_defaults,
                                                   attr="html_column_styles",
                                                   default=html_column_styles_default)
    html_text_styles_default = set_attr_resolver(instance=html_defaults,
                                                 attr="html_text_styles",
                                                 default=html_text_styles_default)

    ####################################################################################################################

    html_frame_styles = set_attr_resolver(instance=html_input,
                                          attr="html_frame_styles",
                                          default=None)
    html_column_styles = set_attr_resolver(instance=html_input,
                                           attr="html_column_styles",
                                           default=None)
    html_text_styles = set_attr_resolver(instance=html_input,
                                         attr="html_text_styles",
                                         default=None)

    ####################################################################################################################

    html_frame_styles = html_frame_mapper(html_frame_input=html_frame_styles,
                                          html_frame_defaults=html_frame_styles_default,
                                          base_frame_defaults=frame_styles,
                                          html_px_multiplier=html_px_multiplier)
    html_column_styles = html_column_mapper(html_columns_input=html_column_styles,
                                            html_column_defaults=html_column_styles_default,
                                            base_column_defaults=column_styles)
    html_text_styles = html_text_mapper(html_text_input=html_text_styles,
                                        html_text_defaults=html_text_styles_default,
                                        base_text_defaults=text_styles,
                                        html_px_multiplier=html_px_multiplier)

    # html_frame_styles = html_frame_mapper(html_frame_input=html_frame_styles,
    #                                       html_frame_defaults=html_frame_styles_default,
    #                                       html_px_multiplier=html_px_multiplier)
    # html_column_styles = html_column_mapper(html_columns_input=html_column_styles,
    #                                         html_column_defaults=html_column_styles_default)
    # html_text_styles = html_text_mapper(html_text_input=html_text_styles,
    #                                     html_text_defaults=html_text_styles_default,
    #                                     html_px_multiplier=html_px_multiplier)

    test = HtmlFrameStyles()

    ####################################################################################################################
    ####################################################################################################################
    ####################################################################################################################

    return FrameStore(frame_styles=frame_styles,
                      column_styles=column_styles,
                      text_styles=text_styles,
                      html_frame_styles=html_frame_styles,
                      html_column_styles=html_column_styles,
                      html_text_styles=html_text_styles)



import math
from typing import List, Optional

from tablate.library.formatters.console.ascii_styler import ascii_background_styler, \
    ascii_text_styler, ascii_terminator
from tablate.library.formatters.console.concat_string import concat_string
from tablate.type.type_input import BaseColumnInput
from tablate.type.type_style import ColumnStyles


def cell_string_single_line(string: str,
                            column_item: BaseColumnInput,
                            column_styles: ColumnStyles,
                            trunc_value: str = "...") -> str:
    string = str(string)
    string = string.split("\n")[0]
    string = " ".join(string.split("\t"))

    padding = column_item["padding"]

    if hasattr(column_styles, "background_padding") and "background" in column_item \
            and column_styles.background_padding is not None and column_item["background"] is not None:
        padding = padding + column_styles.background_padding

    string, white_space = concat_string(string=string,
                                        width=column_item["width"],
                                        padding=padding,
                                        trunc_value=trunc_value)

    if hasattr(column_styles, "background_padding") and "background" in column_item \
            and column_styles.background_padding is not None and column_item["background"] is not None:
        string = f"{' ' * column_styles.background_padding}{string}{' ' * column_styles.background_padding}"

    return_string = ""

    if column_item["text_color"] is not None or column_item["text_style"] is not None:
        if len(string.strip()) > 0:
            fore_space = len(string) - len(string.lstrip())
            back_space = len(string) - len(string.rstrip())
            string = (" " * fore_space) + ascii_text_styler(string=string.strip(), column_dict=column_item) + (
                        " " * back_space)

    if column_item["text_align"] == "left":
        return_string += string
        return_string += " " * white_space
    if column_item["text_align"] == "center":
        if white_space % 2 != 0:
            left_space = int(math.floor(white_space / 2))
            right_space = int(math.floor(white_space / 2) + 1)
        else:
            left_space = int(white_space / 2)
            right_space = int(white_space / 2)
        return_string += " " * left_space
        return_string += string
        return_string += " " * right_space
    if column_item["text_align"] == "right":
        return_string += " " * white_space
        return_string += string

    if column_item["background"] is not None:
        return_string = ascii_background_styler(string=return_string, column_dict=column_item)
    padding = " " * column_item["padding"]
    return_string = f"{padding}{return_string}{ascii_terminator()}{padding}"

    return return_string


def cell_string_multi_line(string: str,
                           column_item: BaseColumnInput,
                           column_styles: ColumnStyles,
                           max_lines: Optional[int] = None,
                           trunc_value: str = "...") -> List[str]:
    return_string_array = []

    string = str(string)

    clean_string = " ".join(string.split("\t"))
    initial_string_array = clean_string.split("\n")

    reached_max_lines = False

    for initial_string_item in initial_string_array:
        word_string_array = initial_string_item.split(" ")
        character_count = 0
        current_line_array = []
        for word_string_index, word_string_item in enumerate(word_string_array):
            last_line = max_lines is not None and len(return_string_array) == max_lines - 1
            character_count += len(word_string_item)
            if (character_count + (column_item["padding"] * 2) < column_item["width"]) or (
                    last_line and character_count + (column_item["padding"] * 2) < column_item["width"] + len(
                    trunc_value)):
                current_line_array.append(word_string_item)
            else:
                if last_line:
                    current_line_array.append(word_string_item)
                    current_line_string = " ".join(current_line_array)
                    return_string_array.append(cell_string_single_line(string=current_line_string,
                                                                       column_item=column_item,
                                                                       column_styles=column_styles,
                                                                       trunc_value=trunc_value))
                    reached_max_lines = True
                    break
                else:
                    current_line_string = " ".join(current_line_array)
                    return_string_array.append(cell_string_single_line(string=current_line_string,
                                                                       column_item=column_item,
                                                                       column_styles=column_styles,
                                                                       trunc_value=trunc_value))
                    current_line_array = [word_string_item]
                    character_count = len(word_string_item)
            character_count += 1
        if reached_max_lines:
            break
        current_line_string = " ".join(current_line_array)
        return_string_array.append(cell_string_single_line(string=current_line_string,
                                                           column_item=column_item,
                                                           column_styles=column_styles,
                                                           trunc_value=trunc_value))
    return return_string_array



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import RowsStylesInput
from tablate.type.type_style import TableRowsStyles


def base_row_input_to_style(row_styles_input: RowsStylesInput) -> TableRowsStyles:

    row_styles_input = row_styles_input if row_styles_input is not None else {}

    row_line_divider = set_key_resolver(instance=row_styles_input, key="row_line_divider", default=None)
    odds_background = set_key_resolver(instance=row_styles_input, key="odds_background", default=None)
    evens_background = set_key_resolver(instance=row_styles_input, key="evens_background", default=None)

    return TableRowsStyles(
        row_line_divider=row_line_divider,
        odds_background=odds_background,
        evens_background=evens_background
    )



from typing import List, Union

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.initializers.processors.process_frame import process_frame
from tablate.type.primitives import FrameDivider, ColumnDivider, OuterBorder, OuterPadding, OuterWidth, Background, \
    BackgroundPadding, HtmlPxMultiplier, Multiline, MaxLines, ColumnPadding, TextStyle, TextAlign, TextColor, FrameName, \
    HtmlDefaultColors
from tablate.type.type_input import GridColumnInput, HtmlOuterStylesInput, HtmlGridStylesInput


class Grid(TablateApiItem):

    def __init__(self,
                 # TablateGrid args
                 columns: List[Union[str, GridColumnInput]],
                 name: FrameName = None,
                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,
                 column_divider: ColumnDivider = None,
                 column_padding: ColumnPadding = None,
                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,
                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlGridStylesInput = None,
                 # TablateApi args
                 html_default_colors: HtmlDefaultColors = None,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                outer_border=outer_border,
                                outer_padding=outer_padding,
                                frame_divider=frame_divider,
                                outer_width=outer_width,
                                html_outer_styles=html_outer_styles,
                                html_default_colors=html_default_colors)

        frame_dict = process_frame(frame_name=name,
                                   frame_type="grid",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict



from __future__ import annotations

from importlib.metadata import version

from tablate.api.Tablate import Tablate
from tablate.api.modules.Grid import Grid
from tablate.api.modules.Table import Table
from tablate.api.modules.Text import Text

from tablate.api.functions.concat import concat

# from tablate.type.type_input import HtmlOuterStylesInput, FrameStylesInput, HtmlFrameStylesInput, ColumnStylesInput, \
#     TextStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput, RowsStylesInput, HtmlRowsStylesInput, \
#     HtmlColumnInput, BaseColumnInput, GridColumnInput, TableColumnInput, TableRowsInput, BaseStylesInput, \
#     TableHeaderFrameStylesInput, TableBodyFrameStylesInput, HtmlStylesInput, HtmlTextFrameStylesInput, \
#     HtmlGridStylesInput, HtmlTableStylesInput, HtmlTableHeaderStylesInput, HtmlTableBodyStylesInput

########################################################################################################################
########################################################################################################################
########################################################################################################################


try:
    __version__ = version('tablate')
except ModuleNotFoundError:
    pass



from typing import Literal, Union, Optional, Tuple, List, Type

Colors = Optional[Literal[
    # "inverted",
    "white",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "dark_red",
    "dark_green",
    "dark_yellow",
    "dark_blue",
    "dark_magenta",
    "dark_cyan",
    "dark_grey",
    "black",
]]

# todo: implement inverted colour

FrameName = str
FrameType = Literal["text", "grid", "table"]

FrameArgs = dict
GlobalArgs = dict

TextString = Union[str, int, float]
TableRowKey = str

TextAlign = Optional[Literal["left", "center", "right"]]
TextStyle = Optional[Literal["normal", "bold", "underlined", "bold_underlined"]]
TextColor = Colors

HeaderAlignment = Literal["column", "left", "center", "right"]

OuterBorder = Literal["blank", "thin", "thick", "double"]
FrameDivider: Type[str] = Literal["none", "blank", "thin", "thick", "double"]
ColumnDivider = Literal["blank", "thin", "thick", "double"]

HtmlDefaultColors = bool

HeaderColumnDivider = Literal["rows", "blank", "thin", "thick", "double"]

HideHeader = bool

CharInteger = int
PercentString = str

ColumnWidth = Optional[Union[CharInteger, PercentString]]
ColumnPadding = CharInteger
Background = Colors
BackgroundPadding = int

MaxLines = Optional[int]
Multiline = Optional[bool]
TruncValue = str

OuterPadding = CharInteger
OuterWidth = CharInteger

CssStyleBlock = str
CssString = str
CssColor = str
PxInteger = int

HtmlPxMultiplier = int

# HtmlBorder = Tuple[Literal["none", "solid", "double"], PxInteger, CssColor]
HtmlBorder = Literal["none", "thin", "thick", "double"]
HtmlSpacer = Union[PxInteger, List[int]]

HtmlTextStyle = Optional[Literal["normal", "bold", "italic"]]
HtmlTextAlign = Optional[Literal["left", "center", "right", "justify"]]
HtmlTextColor = Optional[CssColor]
HtmlTextSize = Optional[PxInteger]

HtmlColumnWidth = Optional[PercentString]

HtmlDividerWeight = Optional[PxInteger]
HtmlDividerColor = Optional[CssString]
HtmlColumnDividerStyle = Optional[HtmlBorder]
HtmlColumnPadding = Optional[HtmlSpacer]
HtmlBackground = Optional[CssString]

HtmlFrameDivider = Optional[HtmlBorder]

HtmlOuterBorder = Optional[HtmlBorder]
HtmlOuterPadding = Optional[HtmlSpacer]
HtmlOuterWidth = Optional[PercentString]  # this will be a calc % - padding

HtmlFrameType = Literal["head", "body"]

HtmlRowGroupElement = Literal["tbody", "thead"]

HtmlCellElement = Literal["td", "th"]



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import HtmlRowsStylesInput
from tablate.type.type_style import HtmlTableRowsStyles


def html_row_input_to_style(html_row_styles_input: HtmlRowsStylesInput) -> HtmlTableRowsStyles:

    html_row_styles_input = html_row_styles_input if html_row_styles_input is not None else {}

    html_row_line_divider_weight = set_key_resolver(instance=html_row_styles_input, key="html_row_line_divider_weight", default=None)
    html_row_line_divider_style = set_key_resolver(instance=html_row_styles_input, key="html_row_line_divider_style", default=None)
    html_row_line_divider_color = set_key_resolver(instance=html_row_styles_input, key="html_row_line_divider_color", default=None)
    html_odds_background = set_key_resolver(instance=html_row_styles_input, key="html_odds_background", default=None)
    html_evens_background = set_key_resolver(instance=html_row_styles_input, key="html_evens_background", default=None)

    return HtmlTableRowsStyles(
        html_row_line_divider_weight=html_row_line_divider_weight,
        html_row_line_divider_style=html_row_line_divider_style,
        html_row_line_divider_color=html_row_line_divider_color,
        html_odds_background=html_odds_background,
        html_evens_background=html_evens_background
    )


from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import text_style_default, text_align_default, text_color_default
from tablate.type.type_style import TextStyles


def base_text_mapper(text_input: TextStyles = None,
                     text_defaults: TextStyles = None) -> TextStyles:
    ####################################################################################################################

    text_input = text_input if text_input is not None else TextStyles()
    text_defaults = text_defaults if text_defaults is not None else TextStyles()

    ####################################################################################################################

    text_return = TextStyles(text_style=set_attr_resolver(instance=text_input,
                                                          attr="text_style",
                                                          default=set_attr_resolver(instance=text_defaults,
                                                                                    attr="text_style",
                                                                                    default=text_style_default)),
                             text_align=set_attr_resolver(instance=text_input,
                                                          attr="text_align",
                                                          default=set_attr_resolver(instance=text_defaults,
                                                                                    attr="text_align",
                                                                                    default=text_align_default)),
                             text_color=set_attr_resolver(instance=text_input,
                                                          attr="text_color",
                                                          default=set_attr_resolver(instance=text_defaults,
                                                                                    attr="text_color",
                                                                                    default=text_color_default)))

    return text_return



from typing import Dict

from tablate.type.primitives import Colors

ascii_character_colors: Dict[Colors, str] = {
    "default": "",
    "white": "97",
    "grey": "37",
    "red": "91",
    "green": "92",
    "yellow": "93",
    "blue": "94",
    "magenta": "95",
    "cyan": "96",
    "dark_red": "31",
    "dark_green": "32",
    "dark_yellow": "33",
    "dark_blue": "34",
    "dark_magenta": "35",
    "dark_cyan": "36",
    "dark_grey": "90",
    "black": "30",
}



from typing import List

from tablate.classes.options.html.style.utilities.base_values import styles_key


def css_factory(style_dict: dict) -> str:

    # def recurse_style_dict_original(style_object: dict, classname_list=None, return_list=None, i=0) -> List[str]:
    #     if return_list is None:
    #         return_list = []
    #     if classname_list is None:
    #         classname_list = []
    #     i += 1
    #     if type(style_object) == dict:
    #         for key, value in style_object.items():
    #             if key == styles_key:
    #             else:
    #                 if len(classname_list) >= i:
    #                     classname_list[i - 1] = key
    #                 else:
    #                     classname_list.append(key)
    #             recurse_style_dict(value, classname_list, return_list, i)
    #     if type(style_object) == list:
    #         return_list.append("." + ".".join(classname_list) + "{" + ";".join(style_object) + ";" + "}")
    #     return return_list

    # Could use depth to clear out previous elements => works but ugly...

    def recurse_style_dict(style_object: dict, classname_list=None, return_list=None, i=0) -> List[str]:
        if return_list is None:
            return_list = []
        if classname_list is None:
            classname_list = []
        i += 1
        if type(style_object) == dict:
            for key, value in style_object.items():
                if key != styles_key:
                    if len(classname_list) >= i:
                        classname_list[i - 1] = key
                    else:
                        classname_list = classname_list[0:i]
                        classname_list.append(key)
                recurse_style_dict(value, classname_list, return_list, i)
        if type(style_object) == list:
            normal_styles = []
            pseudo_styles = {}
            for style_item in style_object:
                if style_item["pseudo"] is not None:
                    if style_item["pseudo"] in pseudo_styles:
                        pseudo_styles[style_item["pseudo"]].append(style_item["style"])
                    else:
                        pseudo_styles[style_item["pseudo"]] = [style_item["style"]]
                else:
                    normal_styles.append(style_item["style"])
            if len(normal_styles) > 0:
                return_list.append("." + ".".join(classname_list) + "{" + ";".join(normal_styles) + ";" + "}")
            for pseudo_style_key, pseudo_style_value in pseudo_styles.items():
                return_list.append("." + ".".join(classname_list) + pseudo_style_key + "{" + ";".join(pseudo_style_value) + ";" + "}")
        return return_list

    return "".join(recurse_style_dict(style_dict))

# todo: FIX THIS STEAMING MESS!!! try to implement proper recursion...
# todo: update:: fixed a bug where previous branches of style object tree were being left intact...
#  ugly slice copy solution => this needs revisiting with a timer to figure out the best solution
# t



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import ColumnStylesInput
from tablate.type.type_style import ColumnStyles


def base_column_input_to_style(column_styles_input: ColumnStylesInput) -> ColumnStyles:

    column_styles_input = column_styles_input if column_styles_input is not None else {}

    column_divider = set_key_resolver(instance=column_styles_input, key="column_divider", default=None)
    padding = set_key_resolver(instance=column_styles_input, key="padding", default=None)
    background_padding = set_key_resolver(instance=column_styles_input, key="background_padding", default=None)

    return ColumnStyles(
        column_divider=column_divider,
        padding=padding,
        background_padding=background_padding
    )



from shutil import get_terminal_size

from tablate.type.primitives import Colors, TextAlign, TextStyle, TextColor, HeaderAlignment, OuterBorder, \
    ColumnDivider, HeaderColumnDivider, HideHeader, ColumnWidth, ColumnPadding, Background, FrameDivider, \
    MaxLines, Multiline, TruncValue, OuterPadding, OuterWidth, HtmlTextStyle, HtmlTextAlign, \
    HtmlTextColor, HtmlTextSize, HtmlColumnDividerStyle, HtmlColumnPadding, HtmlBackground, \
    HtmlFrameDivider, HtmlPxMultiplier, HtmlOuterBorder, HtmlOuterPadding, HtmlOuterWidth, FrameName, HtmlColumnWidth, \
    HtmlDefaultColors, HtmlDividerWeight

global_text_align_default: TextAlign = None
global_text_style_default: TextStyle = None
global_multiline_default: Multiline = None

############################################################################################

frame_name_default: FrameName = ""

colors_default: Colors = None

text_align_default: TextAlign = "left"
text_style_default: TextStyle = "normal"
text_color_default: TextColor = None

outer_padding_default: OuterPadding = 1
outer_width_default: OuterWidth = get_terminal_size((120 + (outer_padding_default * 2), 0))[0] - (outer_padding_default * 2)

outer_border_default: OuterBorder = "thick"
frame_divider_default: FrameDivider = "thick"
column_divider_default: ColumnDivider = "thin"

html_default_colors_default: HtmlDefaultColors = True

column_width_default: ColumnWidth = None
column_padding_default: ColumnPadding = 1
background_default: Background = None
background_padding_default = 1

max_lines_default: MaxLines = None
multiline_default: Multiline = True
trunc_value_default: TruncValue = "..."

############################################################################################

header_base_divider_default = "thin"

table_multiline_default: Multiline = False

table_header_text_style_default: TextStyle = "bold"
table_header_text_align_default: HeaderAlignment = "column"
table_header_column_divider_default: HeaderColumnDivider = "rows"

row_line_divider_default: FrameDivider = "thin"

hide_header_default: HideHeader = False

############################################################################################

html_padding_default = 6
html_divider_weight_default = 1
html_divider_color_default = 'black'
html_px_multiplier_default: HtmlPxMultiplier = 1
html_background_default: HtmlBackground = background_default

html_outer_width_default: HtmlOuterWidth = "100%"
html_outer_border_default: HtmlOuterBorder = outer_border_default
html_outer_padding_default: HtmlOuterPadding = html_padding_default

html_frame_divider_default: HtmlFrameDivider = frame_divider_default

html_column_width_default: HtmlColumnWidth = None
html_column_divider_style_default: HtmlColumnDividerStyle = column_divider_default
html_column_divider_weight_default: HtmlDividerWeight = html_divider_weight_default
html_column_padding_default: HtmlColumnPadding = column_padding_default

html_text_style_default: HtmlTextStyle = text_style_default
html_text_align_default: HtmlTextAlign = text_align_default
html_text_color_default: HtmlTextColor = text_color_default
html_text_size_default = 16

html_max_lines_default: MaxLines = max_lines_default
html_multiline_default: Multiline = multiline_default

html_row_line_divider_style_default: HtmlFrameDivider = row_line_divider_default



from copy import deepcopy
from typing import Union, Tuple

from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.library.initializers.grid_init import grid_init
from tablate.library.initializers.table_init import table_init
from tablate.library.initializers.text_init import text_init
from tablate.type.primitives import FrameType, FrameArgs, FrameName
from tablate.type.type_global import Globals
from tablate.type.type_store import GridFrameStore, TableHeaderFrameStore, TableBodyFrameStore, FrameDictList, FrameDict


def process_frame_args(frame_args: FrameArgs,
                       frame_type: FrameType,
                       global_options: Globals) -> Union[GridFrameStore, Tuple[Union[TableHeaderFrameStore, None], TableBodyFrameStore]]:

    frame_args = deepcopy(frame_args)

    # if frame_type == "text":
    #     return text_init(**frame_args, global_options=global_options)
    if frame_type == "grid" or frame_type == "text":
        return grid_init(**frame_args, global_options=global_options)
    if frame_type == "table":
        return table_init(**frame_args, global_options=global_options)


def process_frame(frame_name: FrameName,
                  frame_type: FrameType,
                  frame_args: FrameArgs,
                  frame_list: FrameDictList,
                  global_options: Globals) -> FrameDict:
    if "self" in frame_args:
        del frame_args["self"]
    if "html_default_colors" in frame_args:
        del frame_args["html_default_colors"]
    if "outer_border" in frame_args:
        del frame_args["outer_border"]
    if "outer_padding" in frame_args:
        del frame_args["outer_padding"]
    if "outer_width" in frame_args:
        del frame_args["outer_width"]
    if "html_outer_styles" in frame_args:
        del frame_args["html_outer_styles"]

    if frame_type == "text":
        frame_args["columns"] = [{"string": frame_args["text"]}]
        del frame_args["text"]
        frame_args["column_padding"] = frame_args["frame_padding"]
        del frame_args["frame_padding"]
        frame_args["column_divider"] = "thin"

    if frame_type == "grid" and type(frame_args["columns"][0]) == str:
        grid_columns = []
        for column_item in frame_args["columns"]:
            grid_columns.append({
                "string": column_item
            })
        frame_args["columns"] = grid_columns

    name = gen_frame_name(name=frame_name, type=frame_type, frame_dict=frame_list)
    args = deepcopy(frame_args)
    args["name"] = name

    return FrameDict(name=name, type=frame_type, args=args, store=process_frame_args(frame_args=args,
                                                                                     frame_type=frame_type,
                                                                                     global_options=global_options))



from copy import copy
from typing import Callable, Union, List, Tuple, Optional

from tablate.classes.options.html.style.mixins.AddStyleMixin import AddStyleMixin
from tablate.classes.options.html.style.mixins.ClassNameMixin import ClassNameMixin
from tablate.classes.options.html.style.subclasses.TextStyler import TextStyler
from tablate.classes.options.html.style.utilities.selectors import column_selector, row_selector
from tablate.classes.options.html.style.utilities.style_types import ElementSelectorDict, SelectorDictUnion, \
    ElementSelectorDictKeys, StyleItem


class ElementStyler(ClassNameMixin, AddStyleMixin):

    def __init__(self,
                 selector: ElementSelectorDict,
                 create_style: Callable[[SelectorDictUnion, StyleItem], None]) -> None:
        self._selector_dict = selector
        self._create_style = create_style

    def column(self, column_index: int = None):
        selector = self.__create_selector_dict(column=True, column_index=column_index)
        return ElementStyler(selector=selector, create_style=self._create_style)

    def row(self, row_index: int = None):
        selector = self.__create_selector_dict(row=True, row_index=row_index)
        return ElementStyler(selector=selector, create_style=self._create_style)

    @property
    def text(self):
        selector = self.__create_selector_dict(text_cell=True)
        return TextStyler(selector=selector, create_style=self._create_style)

    def __create_selector_dict(self,
                               column: bool = False,
                               column_index: int = None,
                               row: bool = None,
                               row_index: int = None,
                               text_cell: bool = False) -> ElementSelectorDict:
        new_selector_dict: ElementSelectorDict = copy(self._selector_dict)
        if column:
            # base_type: ElementSelectorDictKeys = "tablate_column"
            new_selector_dict["element_type"]: ElementSelectorDictKeys = "tablate_column"
            new_selector_dict["tablate_column"] = column_selector(index=column_index)
        if row:
            # base_type: ElementSelectorDictKeys = "tablate_row"
            new_selector_dict["element_type"]: ElementSelectorDictKeys = "tablate_row"
            new_selector_dict["tablate_row"] = row_selector(index=row_index)
        if text_cell:
            # base_type: ElementSelectorDictKeys = "tablate_text"
            new_selector_dict["element_type"]: ElementSelectorDictKeys = "tablate_text"
        return new_selector_dict



from typing import Union

from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.library.initializers.mappers.object.element.base_column_input_to_style import base_column_input_to_style
from tablate.library.initializers.mappers.object.element.base_frame_input_to_style import base_frame_input_to_style
from tablate.library.initializers.mappers.object.element.base_row_input_to_style import base_row_input_to_style
from tablate.library.initializers.mappers.object.element.base_text_input_to_style import base_text_input_to_style
from tablate.type.type_input import BaseStylesInput, TableBodyFrameStylesInput
from tablate.type.type_store import BaseFrameStore, TableFrameStore, FrameStore


def base_frame_input_to_store(frame_input: Union[BaseStylesInput, TableBodyFrameStylesInput],
                              frame_type: str) -> BaseFrameStore:

    if frame_type == "table":
        return TableFrameStore(
            frame_styles=base_frame_input_to_style(
                set_key_resolver(instance=frame_input, key="frame_styles",
                                 default={})),
            column_styles=base_column_input_to_style(
                set_key_resolver(instance=frame_input, key="column_styles",
                                 default={})),
            row_styles=base_row_input_to_style(
                set_key_resolver(instance=frame_input, key="row_styles",
                                 default={})),
            text_styles=base_text_input_to_style(
                set_key_resolver(instance=frame_input, key="text_styles",
                                 default={})))
    else:
        return FrameStore(
            frame_styles=base_frame_input_to_style(
                set_key_resolver(instance=frame_input, key="frame_styles",
                                 default={})),
            column_styles=base_column_input_to_style(
                set_key_resolver(instance=frame_input, key="column_styles",
                                 default={})),
            text_styles=base_text_input_to_style(
                set_key_resolver(instance=frame_input, key="text_styles",
                                 default={})))



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import HtmlTextStylesInput
from tablate.type.type_style import HtmlTextStyles


def html_text_input_to_style(html_text_styles_input: HtmlTextStylesInput) -> HtmlTextStyles:

    html_text_styles_input = html_text_styles_input if html_text_styles_input is not None else {}

    html_text_style = set_key_resolver(instance=html_text_styles_input, key="html_text_style", default=None)
    html_text_align = set_key_resolver(instance=html_text_styles_input, key="html_text_align", default=None)
    html_text_color = set_key_resolver(instance=html_text_styles_input, key="html_text_color", default=None)
    html_text_size = set_key_resolver(instance=html_text_styles_input, key="html_text_size", default=None)

    return HtmlTextStyles(
        html_text_style=html_text_style,
        html_text_align=html_text_align,
        html_text_color=html_text_color,
        html_text_size=html_text_size
    )



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import TextStylesInput
from tablate.type.type_style import TextStyles


def base_text_input_to_style(text_input: TextStylesInput) -> TextStyles:

    text_input = text_input if text_input is not None else {}

    text_style = set_key_resolver(instance=text_input, key="text_style", default=None)
    text_align = set_key_resolver(instance=text_input, key="text_align", default=None)
    text_color = set_key_resolver(instance=text_input, key="text_color", default=None)

    return TextStyles(
        text_style=text_style,
        text_align=text_align,
        text_color=text_color
    )



from copy import deepcopy
from typing import List, Union

from tablate.classes.bases.TablateApiSet import TablateApiSet
from tablate.library.initializers.processors.process_frame import process_frame
from tablate.type.primitives import FrameDivider, Multiline, TextColor, \
    TextAlign, TextStyle, MaxLines, HideHeader, Background, ColumnDivider, ColumnPadding, HtmlPxMultiplier, \
    BackgroundPadding, FrameName, OuterBorder, OuterPadding, OuterWidth, HtmlDefaultColors
from tablate.type.type_input import GridColumnInput, TableColumnInput, TableRowsInput, HtmlOuterStylesInput, \
    ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput, HtmlTextFrameStylesInput, HtmlGridStylesInput, \
    HtmlTableHeaderStylesInput, HtmlTableBodyStylesInput, HtmlTableStylesInput, \
    TableHeaderFrameStylesInput, TableBodyFrameStylesInput


class Tablate(TablateApiSet):
    """
    Container for text, grid and table frames. Allows cumulative addition of frames to an instance.

     Note:
         When using the 'Tablate' class in an IPython environment, it is important to name each frame (or else rerunning
            cells will create duplicate entries).

    Args:
        outer_border:
            Outer border style for the Tablate container instance.
                [default: 'thick']
        outer_padding:
            Outer padding (in Unicode character width) for the Tablate container instance.
                [default: 1]
        outer_width:
            Outer width (in Unicode character width) for the Tablate container instance.
                [default: 120 or terminal width]
        html_default_colors:
            Apply the default HTML color scheme to the Tablate instance.
                [default: True]
        frame_divider:
            Default divider between frames within the Tablate container instance.
                [default: 'thick']
        background:
            Default background for frames withing the Tablate container instance.
                [default: None]
        background_padding:
            Additional padding applied to cells with a background color set when outputting to ASCII.
                [default: 1]
        html_px_multiplier:
            Multiplier for HTML properties inherited from base (ASCII) styling.
                [default: 6]
        html_outer_styles:
            Style dictionary for HTML specific styles for Tablate container instance.
                [default: None]
        column_styles:
            Style dictionary for default column styles.
                [default: {"column_divider": 'thin', "padding": 1, "background_padding": 1}]
        text_styles:
            Style dictionary for default text styles.
                [default: {
                    "text_style": 'normal',
                    "text_align": 'left',
                    "text_color": None
                    }]
        html_frame_styles:
            Style dictionary for default HTML frame styles.
                [default: {
                    "html_frame_divider_style": "thick",
                    "html_frame_divider_weight: 1,
                    "html_frame_divider_color": 'black',
                    "html_multiline": True (False for Table frames),
                    "html_max_lines": None,
                    "html_background": None
                    }]
        html_column_styles:
            Style dictionary for default HTML column styles.
                [default: {
                    "html_column_divider_style": 'thin',
                    "html_column_divider_weight": 1,
                    "html_column_divider_color": 'black',
                    "html_padding": 6px
                    }]
        html_text_styles:
            Style dictionary for default HTML text styles.
                [default: {
                    "html_text_style": 'normal',
                    "html_text_align": 'left',
                    "html_text_color": None,
                    "html_text_size": 16
                    }]
    """

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 html_default_colors: HtmlDefaultColors = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        args = deepcopy(locals())
        del args["self"]

        TablateApiSet.__init__(self=self, **args)

    def add_text_frame(self,
                       text: Union[str, int, float],

                       name: FrameName = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       frame_divider: FrameDivider = None,
                       frame_padding: ColumnPadding = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlTextFrameStylesInput = None) -> None:
        """
        Add a new text frame (single row, single column) to Tablate container instance.

        Args:
            text: Text frame content.
            name: Frame name (to prevent duplicate entries). [default: None]
            text_style: [default: 'normal']
            text_align: [default: 'left']
            text_color: [default: 'None']
            frame_divider: Dividing line at the bottom of the frame. [default: 'thick']
            frame_padding: Text padding within Text frame. [default: 1]
            background: Background color for the Text frame. [default: None]
            background_padding: Additional padding applied to text if background color set. [default: 1]
            multiline: Whether the text may wrap to multiple lines. [default: True]
            max_lines: Maximum number of lines to which the text may wrap. [default: None]
            html_px_multiplier: Multiplier for HTML properties inherited from base (ASCII) styling. [default: 6]
            html_styles: {"html_frame_styles": {"html_frame_divider_style": "thick", "html_frame_divider_weight: 1, "html_frame_divider_color": 'black', "html_multiline": True (False for Table frames), "html_max_lines": None, "html_background": None}, "html_text_styles": {"html_text_style": 'normal', "html_text_align": 'left', "html_text_color": None, "html_text_size": 16}}]
        """
        frame_dict = process_frame(frame_name=name,
                                   frame_type="text",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict

    def add_grid_frame(self,
                       columns: List[Union[str, GridColumnInput]],

                       name: FrameName = None,

                       frame_divider: FrameDivider = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       column_divider: ColumnDivider = None,
                       column_padding: ColumnPadding = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlGridStylesInput = None) -> None:
        """

        Add a new grid frame (single row of multiple columns) to Tablate instance.

        Args:
            columns:
            name:
            frame_divider:
            background:
            background_padding:
            multiline:
            max_lines:
            column_divider:
            column_padding:
            text_style:
            text_align:
            text_color:
            html_px_multiplier:
            html_styles:
        """
        frame_dict = process_frame(frame_name=name,
                                   frame_type="grid",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict

    def add_table_frame(self,
                        columns: List[TableColumnInput],
                        rows: List[TableRowsInput],

                        name: FrameName = None,

                        frame_divider: FrameDivider = None,
                        multiline: Multiline = None,
                        max_lines: MaxLines = None,
                        background: Background = None,
                        background_padding: BackgroundPadding = None,

                        multiline_header: Multiline = None,
                        max_lines_header: MaxLines = None,
                        hide_header: HideHeader = None,

                        column_divider: ColumnDivider = None,
                        column_padding: ColumnPadding = None,
                        header_base_divider: FrameDivider = None,

                        row_line_divider: FrameDivider = None,
                        odd_row_background: Background = None,
                        even_row_background: Background = None,

                        text_style: TextStyle = None,
                        text_align: TextAlign = None,
                        text_color: TextColor = None,

                        header_styles: TableHeaderFrameStylesInput = None,
                        body_styles: TableBodyFrameStylesInput = None,

                        html_px_multiplier: HtmlPxMultiplier = None,
                        html_styles: HtmlTableStylesInput = None,

                        html_header_styles: HtmlTableHeaderStylesInput = None,
                        html_body_styles: HtmlTableBodyStylesInput = None) -> None:
        """
        Add a new table frame (multiple rows of multiple columns) to Tablate instance.

        Args:
            columns:
            rows:
            name:
            frame_divider:
            multiline:
            max_lines:
            background:
            background_padding:
            multiline_header:
            max_lines_header:
            hide_header:
            column_divider:
            column_padding:
            header_base_divider:
            row_line_divider:
            odd_row_background:
            even_row_background:
            text_style:
            text_align:
            text_color:
            header_styles:
            body_styles:
            html_px_multiplier:
            html_styles:
            html_header_styles:
            html_body_styles:
        """
        frame_dict = process_frame(frame_name=name,
                                   frame_type="table",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict



from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.initializers.processors.process_frame import process_frame_args
from tablate.type.primitives import OuterBorder, OuterPadding, OuterWidth, FrameDivider, Background, BackgroundPadding, \
    HtmlPxMultiplier, HtmlDefaultColors
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import FrameDict


class TablateItem(TablateApiItem):

    def __init__(self,
                 frame_item: FrameDict,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 html_default_colors: HtmlDefaultColors = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None):
        super().__init__(outer_border=outer_border,
                         outer_padding=outer_padding,
                         outer_width=outer_width,
                         frame_divider=frame_divider,
                         html_default_colors=html_default_colors,
                         background=background,
                         background_padding=background_padding,
                         html_px_multiplier=html_px_multiplier,
                         html_outer_styles=html_outer_styles,
                         column_styles=column_styles,
                         text_styles=text_styles,
                         html_frame_styles=html_frame_styles,
                         html_column_styles=html_column_styles,
                         html_text_styles=html_text_styles)

        self._frame_list = {frame_item.name: FrameDict(name=frame_item.name,
                                                       type=frame_item.type,
                                                       args=frame_item.args,
                                                       store=process_frame_args(frame_args=frame_item.args,
                                                                                frame_type=frame_item.type,
                                                                                global_options=self._globals_store.store))}



from copy import deepcopy
from typing import List

from tablate.classes.bases.TablateApiSet import TablateApiSet
from tablate.library.calcs.gen_frame_name import gen_frame_name
from tablate.type.primitives import OuterBorder, OuterPadding, OuterWidth, FrameDivider, Background, BackgroundPadding, \
    HtmlPxMultiplier
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import FrameDictList


class TablateSet(TablateApiSet):

    def __init__(self,
                 frame_list: List[FrameDictList],
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None):

        super().__init__(outer_border=outer_border,
                         outer_padding=outer_padding,
                         outer_width=outer_width,
                         frame_divider=frame_divider,
                         background=background,
                         background_padding=background_padding,
                         html_px_multiplier=html_px_multiplier,
                         html_outer_styles=html_outer_styles,
                         column_styles=column_styles,
                         text_styles=text_styles,
                         html_frame_styles=html_frame_styles,
                         html_column_styles=html_column_styles,
                         html_text_styles=html_text_styles)

        for frame_set in frame_list:
            for frame_key, frame_item in frame_set.items():
                name = gen_frame_name(name=frame_key, type=frame_item.type, frame_dict=self._frame_list, ensure_unique=True)
                frame_item.name = name
                frame_item.args["name"] = name
                self._frame_list[name] = deepcopy(frame_item)



from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.html.style.attributes.color import background_color_attr
from tablate.type.defaults import background_default, row_line_divider_default, html_divider_weight_default, \
    html_divider_color_default
from tablate.type.type_style import HtmlTableRowsStyles, TableRowsStyles


def html_rows_mapper(html_rows_input: HtmlTableRowsStyles = None,
                     html_rows_defaults: HtmlTableRowsStyles = None,
                     base_rows_defaults: TableRowsStyles = None) -> HtmlTableRowsStyles:
    html_row_line_divider_weight = set_attr_resolver(
        instance=html_rows_input,
        attr="html_row_line_divider_weight",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            attr="html_row_line_divider_weight",
            default=html_divider_weight_default))
    html_row_line_divider_style = set_attr_resolver(
        instance=html_rows_input,
        attr="html_row_line_divider_style",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            attr="html_row_line_divider_style",
            default=set_attr_resolver(
                instance=base_rows_defaults,
                attr="row_line_divider",
                default=row_line_divider_default)))
    html_row_line_divider_color = set_attr_resolver(
        instance=html_rows_input,
        attr="html_row_line_divider_color",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            attr="html_row_line_divider_color",
            default=html_divider_color_default))
    html_odds_background = set_attr_resolver(
        instance=html_rows_input,
        attr="html_odds_background",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            attr="html_odds_background",
            default=background_color_attr(set_attr_resolver(
                instance=base_rows_defaults,
                attr="odds_background",
                default=background_default))))
    html_evens_background = set_attr_resolver(
        instance=html_rows_input,
        attr="html_evens_background",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            attr="html_evens_background",
            default=background_color_attr(set_attr_resolver(
                instance=base_rows_defaults,
                attr="evens_background",
                default=background_default))))

    rows_return = HtmlTableRowsStyles(html_row_line_divider_weight=html_row_line_divider_weight,
                                      html_row_line_divider_style=html_row_line_divider_style,
                                      html_row_line_divider_color=html_row_line_divider_color,
                                      html_odds_background=html_odds_background,
                                      html_evens_background=html_evens_background)

    return rows_return



from tablate import Table
from tablate.classes.bases.TablateApiSet import TablateSet
from tests.test_objs import column_list1, row_list1

new_table1 = Table(column_list1, row_list1)
new_table2 = Table(column_list1, row_list1)

new_table1.print()

new_set = TablateSet([new_table1, new_table2])
print("###")
new_set.print()

newer_set = TablateSet([new_set])


from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import html_outer_width_default, html_px_multiplier_default, outer_border_default, \
    outer_padding_default, html_divider_weight_default, html_padding_default, \
    html_default_colors_default, html_divider_color_default
from tablate.type.primitives import HtmlPxMultiplier, HtmlDefaultColors
from tablate.type.type_style import HtmlOuterStyles, OuterStyles


def html_outer_mapper(html_outer_input: HtmlOuterStyles = None,
                      html_outer_defaults: HtmlOuterStyles = None,
                      base_outer_defaults: OuterStyles = None,
                      html_default_colors: HtmlDefaultColors = None,
                      html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> HtmlOuterStyles:
    html_outer_border_weight = set_attr_resolver(
        instance=html_outer_input,
        attr="html_outer_border_weight",
        default=set_attr_resolver(
            instance=html_outer_defaults,
            attr="html_outer_border_weight",
            default=html_divider_weight_default))
    html_outer_border_style = set_attr_resolver(
        instance=html_outer_input,
        attr="html_outer_border_style",
        default=set_attr_resolver(
            instance=html_outer_defaults,
            attr="html_outer_border_style",
            default=set_attr_resolver(
                instance=base_outer_defaults,
                attr="outer_border",
                default=outer_border_default)))
    html_outer_border_color = set_attr_resolver(
        instance=html_outer_input,
        attr="html_outer_border_color",
        default=set_attr_resolver(
            instance=html_outer_defaults,
            attr="html_outer_border_color",
            default=html_divider_color_default))
    html_outer_padding = set_attr_resolver(
        instance=html_outer_input,
        attr="html_outer_padding",
        default=set_attr_resolver(
            instance=html_outer_defaults,
            attr="html_outer_padding",
            default=set_attr_resolver(
                instance=base_outer_defaults,
                attr="outer_padding",
                default=outer_padding_default) * html_padding_default))
    html_outer_width = set_attr_resolver(
        instance=html_outer_input,
        attr="html_outer_width",
        default=set_attr_resolver(
            instance=html_outer_defaults,
            attr="html_outer_width",
            default=html_outer_width_default))
    html_default_colors = set_attr_resolver(
        instance=html_outer_input,
        attr="html_default_colors",
        default=html_default_colors if html_default_colors is not None else html_default_colors_default)

    html_outer_return = HtmlOuterStyles(html_outer_border_weight=html_outer_border_weight,
                                        html_outer_border_style=html_outer_border_style,
                                        html_outer_border_color=html_outer_border_color,
                                        html_outer_padding=html_outer_padding,
                                        html_outer_width=html_outer_width,
                                        html_px_multiplier=html_px_multiplier,
                                        html_default_colors=html_default_colors)

    return html_outer_return



from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.library.formatters.html.style.attributes.divider import divider_attr
from tablate.library.formatters.html.style.elements.style_column import style_column
from tablate.library.formatters.html.style.elements.style_text import style_text

from tablate.type.type_store import FrameStore


def style_frame(frame_store: FrameStore, frame_styler: ElementStyler, is_last_element: bool) -> None:

    divider_style = frame_store.html_frame_styles.html_frame_divider_style
    divider_weight = frame_store.html_frame_styles.html_frame_divider_weight
    divider_color = frame_store.html_frame_styles.html_frame_divider_color
    background = frame_store.html_frame_styles.html_background

    frame_styler.add_style_attribute("padding", "0", " tr")
    frame_styler.add_style_attribute("padding", "0", " th")
    frame_styler.add_style_attribute("padding", "0", " td")


    if not is_last_element:
        frame_styler.add_style_attribute("border-bottom",
                                         divider_attr(divider_style=divider_style,
                                                      divider_weight=divider_weight,
                                                      divider_color=divider_color))
    if background is not None:
        frame_styler.add_style_attribute("background-color", background)

    multiline = frame_store.html_frame_styles.html_multiline
    max_lines = frame_store.html_frame_styles.html_max_lines

    text_styler = frame_styler.text

    if multiline is False or max_lines is not None:
        text_styler.add_style_attribute("display", "-webkit-box")
    text_styler.add_style_attribute("-webkit-box-orient", "vertical")
    text_styler.add_style_attribute("overflow", "hidden")
    if multiline is False:
        text_styler.add_style_attribute("-webkit-line-clamp", 1)
    elif max_lines is not None:
        text_styler.add_style_attribute("-webkit-line-clamp", max_lines)

    style_column(column_store=frame_store.html_column_styles,
                 column_styler=frame_styler.column(),
                 html_px_multiplier=frame_store.html_frame_styles.html_px_multiplier)
    style_text(text_store=frame_store.html_text_styles,
               text_styler=text_styler,
               html_px_multiplier=frame_store.html_frame_styles.html_px_multiplier)








from dataclasses import dataclass, field
from typing import Literal, Union, List, Dict, Tuple, Optional

from tablate.type.defaults import frame_name_default
from tablate.type.primitives import TextString, TableRowKey, FrameName, HideHeader
from tablate.type.type_input import GridColumnInput, TableColumnInput
from tablate.type.type_style import ColumnStyles, TextStyles, HtmlTextStyles, FrameStyles, HtmlColumnStyles, \
    HtmlFrameStyles, HtmlTableRowsStyles, TableRowsStyles


########################################################################################################################
# FrameDicts ###########################################################################################################
########################################################################################################################

@dataclass
class BaseFrameStore:
    frame_styles: FrameStyles = field(default_factory=FrameStyles)
    column_styles: ColumnStyles = field(default_factory=ColumnStyles)
    text_styles: TextStyles = field(default_factory=TextStyles)


@dataclass
class BaseTableFrameStore(BaseFrameStore):
    row_styles: TableRowsStyles = field(default_factory=TableRowsStyles)


@dataclass
class HtmlFrameStore:
    html_frame_styles: HtmlFrameStyles = field(default_factory=HtmlFrameStyles)
    html_column_styles: HtmlColumnStyles = field(default_factory=HtmlColumnStyles)
    html_text_styles: HtmlTextStyles = field(default_factory=HtmlTextStyles)


@dataclass
class HtmlTableFrameStore(HtmlFrameStore):
    html_row_styles: HtmlTableRowsStyles = field(default_factory=HtmlTableRowsStyles)


@dataclass
class FrameStore(BaseFrameStore, HtmlFrameStore):
    pass


@dataclass
class TableFrameStore(BaseTableFrameStore, HtmlTableFrameStore):
    pass

# Grid FrameDict #######################################################################################################




@dataclass()
class GridFrameStore(FrameStore):
    type: Union[Literal["grid"], Literal["text"]] = "text"
    name: FrameName = frame_name_default
    column_list: List[GridColumnInput] = field(default_factory=list)


# Table FrameDict ######################################################################################################

@dataclass()
class TableHeaderFrameStore(FrameStore):
    type: Literal["table_header"] = "table_header"
    name: FrameName = frame_name_default
    column_list: List[TableColumnInput] = field(default_factory=list)


@dataclass()
class TableBodyFrameStore(FrameStore):
    type: Literal["table_body"] = "table_body"
    name: FrameName = frame_name_default
    hide_header: HideHeader = False
    column_list: List[TableColumnInput] = field(default_factory=list)
    row_list: List[Dict[TableRowKey, TextString]] = field(default_factory=list)
    row_styles: TableRowsStyles = field(default_factory=TableRowsStyles)
    html_row_styles: HtmlTableRowsStyles = field(default_factory=HtmlTableRowsStyles)


# FrameDict List #######################################################################################################

FrameStoreUnion = Union[GridFrameStore, TableHeaderFrameStore, TableBodyFrameStore]
FrameStoreList = List[FrameStoreUnion]

@dataclass()
class FrameDict:
    name: FrameName
    type: Literal["text", "grid", "table"]
    args: dict
    store: Optional[Union[GridFrameStore, Tuple[Union[TableHeaderFrameStore, None], TableBodyFrameStore]]]


FrameDictList = Dict[str, FrameDict]





########################################################################################################################
########################################################################################################################
########################################################################################################################

#
# column_list: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "50%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# column_list2: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "25%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# row_list = [
#     {
#         "Key One": "Some string value",
#         "Key Two": 4,
#         "Key Three": 9.6,
#         "divider": "thick"
#     },
#     {
#         "Key One": "Some other string value",
#         "Key Two": 5,
#         "Key Three": 4.6
#
#     },
#     {
#         "Key One": "Some final string value",
#         "Key Two": 100,
#         "Key Three": 328.832
#
#     }
# ]
# ########################################################################################################################
# html_global_styles_input = HtmlOuterBase(html_outer_border_style="thick",
#                                          html_outer_padding=16,
#                                          html_outer_width="100%")
# ########################################################################################################################
# frame_styles_input = FrameStylesInput(frame_divider="thick",
#                                       max_lines=None,
#                                       multiline=True,
#                                       background="yellow",
#                                       trunc_value="...")
# column_styles_input = ColumnStylesInput(divider="thick",
#                                         padding=1)
# text_styles_input = TextStylesInput(text_style="bold",
#                                     text_align="left",
#                                     text_color="green")
# ########################################################################################################################
# html_frame_styles_input = HtmlFrameStylesInput(html_frame_divider_style="thick",
#                                                html_max_lines=5,
#                                                html_multiline=True,
#                                                html_background="green")
# html_column_styles_input = HtmlColumnStylesInput(html_divider_style="thin",
#                                                  html_padding=6)
# html_text_styles_input = HtmlTextStylesInput(html_text_style="normal",
#                                              html_text_align="left",
#                                              html_text_color="red",
#                                              html_text_size=16)
# ########################################################################################################################
# table_rows_frame_styles_input = TableBodyFrameStylesInput(frame_styles=frame_styles_input,
#                                                           column_styles=column_styles_input,
#                                                           text_styles=text_styles_input,
#                                                           row_styles=TableRowsBase(row_line_divider="thin",
#                                                                                    odds_background="grey",
#                                                                                    evens_background="green"))
# header_frame_styles_input = TableHeaderFrameStylesInput(frame_styles=frame_styles_input,
#                                                         column_styles=column_styles_input,
#                                                         text_styles=text_styles_input)
# html_header_styles_input = HtmlTableHeaderFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                            html_column_styles=html_column_styles_input,
#                                                            html_text_styles=html_text_styles_input)
# html_rows_styles_input = HtmlTableBodyFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                        html_column_styles=html_column_styles_input,
#                                                        html_text_styles=html_text_styles_input,
#                                                        html_row_styles=HtmlTableRowsBase(
#                                                            html_row_line_divider_style="none"
#                                                        ))
# ########################################################################################################################
#
# start = time.perf_counter_ns()
# test = Tablate()
# # test = Tablate(outer_border="thin",
# #                      outer_padding=6,
# #                      outer_width=120,
# #                      html_outer_styles=html_global_styles_input,
# #                      frame_divider="double",
# #                      background="blue",
# #                      column_styles=column_styles_input,
# #                      text_styles=text_styles_input,
# #                      html_frame_styles=html_frame_styles_input,
# #                      html_column_styles=html_column_styles_input,
# #                      html_text_styles=html_text_styles_input)
#
# test.add_text_frame(text="Some String...",
#                     text_style="bold",
#                     text_align="left",
#                     text_color="red",
#                     frame_divider="thick",
#                     background="blue",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_grid_frame(columns=column_list,
#                     frame_divider="thick",
#                     text_color="white",
#                     text_style="bold_underlined",
#                     background="dark_red",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      odd_row_background="red",
#                      text_color="yellow")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      name="Some Table",
#                      frame_divider="thick",
#                      background="blue",
#                      multiline=True,
#                      max_lines=5,
#                      multiline_header=False,
#                      max_lines_header=None,
#                      hide_header=False,
#
#                      column_divider="thin",
#                      column_padding=1,
#                      header_base_divider="thick",
#                      row_line_divider="thin",
#                      even_row_background="magenta",
#                      odd_row_background="blue",
#
#                      text_style="bold",
#                      text_align="center",
#                      text_color="blue",
#
#                      header_styles=header_frame_styles_input,
#                      body_styles=table_rows_frame_styles_input,
#
#                      html_styles=html_rows_styles_input,
#
#                      html_header_styles=html_header_styles_input,
#                      html_body_styles=html_rows_styles_input)
#
#
# test.add_text_frame(text="Some String...")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list2,
#                      rows=row_list,
#                      hide_header=True)
#
# test.add_grid_frame(columns=column_list, html_px_multiplier=3, text_color="blue")
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      text_color="grey")
#
#
# string = test.to_string()
# html = test.to_html()
# print(string)
# print(html)
#
# test.list_frames()
#
# test.get_frame(9).print()

column_list = [
    {
        "string": "Column One",
        "key": "one",
        "width": "50%",
        "divider": "thick",
    },
    {
        "string": "Column Two",
        "key": "two",
        "text_align": "right",

    },
    {
        "string": "Column Three",
        "key": "three",
        "text_align": "right"
    }
]

row_list = [
    {
        "one": "Some string value",
        "two": 4,
        "three": 9.6,
    },
    {
        "one": "Some other string value",
        "two": 5,
        "three": 4.6

    },
    {
        "one": "Some final string value",
        "two": 100,
        "three": 328.832

    }
]

tab = Tablate(outer_border="thin", html_column_styles={"html_column_divider_color": "green"})

tab.add_table_frame(columns=column_list, rows=row_list, background="red", multiline=True)
tab.add_text_frame("one")
tab.add_table_frame(columns=column_list, rows=row_list, background="green")
tab.add_text_frame("two")
tab.add_table_frame(columns=column_list, rows=row_list, background="blue", html_styles={"html_column_styles": {"html_column_divider_color": "red"}})
tab.add_text_frame("three", multiline=True)
tab.add_table_frame(columns=column_list, rows=row_list, background="cyan", hide_header=True, multiline=True)
tab.add_text_frame("four")
tab.add_table_frame(columns=column_list, rows=row_list, background="yellow")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_table_frame(columns=column_list, rows=row_list, background="magenta", name="Test")
tab.add_grid_frame(columns=["Yo", "Ho", "Ho"], html_styles={"html_column_styles": {"html_column_divider_color": "red"}})
tab.add_grid_frame(columns=column_list)





tab.print()
tab.list_frames()
print(tab.to_html())
tab.get_frame(3).print()
tab.get_frame(3, apply_globals=True).print()
tab.list_frames()
got_frame = tab.get_frame(4, apply_globals=True)
tab.replace_frame(3, got_frame, "cat")
tab.list_frames()

tab.apply_style(".tablate_frame_3.tablate_row", "background-color:pink", ":hover")


print(tab.to_html())

new_frame = concat([tab, got_frame])
tab.remove_frame(0)
tab.print()

print(got_frame.name)
got_frame.rename("lala")
print(got_frame.name)

new_frame2 = concat([tab, got_frame], outer_border="thin")

new_frame2.print()

new_frame2.list_frames()

# for item_key, item_name in new_frame._frame_list.items():
#     print(item_key, item_name["name"])

test_frame = new_frame2.get_frame("Test", apply_globals=True)

test_frame.print()

test_frame.apply(lambda name, frame_type, frame_args, global_args: print(name, frame_type, frame_args, global_args))

print(tab.get_frame(9).to_dict())

import pandas as pd


arrays = [[1, 1, 2, 2], ['red', 'blue', 'red', 'blue']]

mul = pd.MultiIndex.from_arrays(arrays)
df3 = pd.DataFrame({"one": ["sdfsdfsdfs sd  fsd f sd fsdfsdfsdf  sdfsdfff sd f sd fsd fsdf sdf sd f sd f sd f sd f s sd f sdfsdfkjsdf ds f dsf sdfsdfsdf sdf sdfsdfdsfdsf dsfsdfsdf",2,3,4], "two": [1,2,3,4]}, index=mul)
tab.from_dict(df3.to_dict())

tab.print()

print(tab.to_html())


new_text = Text("some text")
new_text.print()

new_grid = Grid(columns=column_list)
new_grid.print()

new_table = Table(columns=column_list, rows=row_list, )
new_table.print()




import copy
import math
from typing import List, TypeVar

from tablate.type.type_store import FrameStoreUnion

T = TypeVar("T")


def calc_column_percent(frame_list: List[FrameStoreUnion], outer_width: int) -> (List[T], List[int]):
    frame_list = copy.deepcopy(frame_list)
    baseline_width_set = set()
    outer_width = outer_width - 1
    for frame_item in frame_list:
        outer_width_total = 0
        for column_item in frame_item.column_list:
            outer_width_total += column_item["width"] + 1
            acc_column_width = math.floor((100 / outer_width) * outer_width_total)
            baseline_width_set.add(math.floor(acc_column_width))
            column_item["baseline_width_percent"] = acc_column_width
    sorted_baseline_array = sorted(baseline_width_set)
    return frame_list, sorted_baseline_array






from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import HtmlFrameStylesInput
from tablate.type.type_style import HtmlFrameStyles


def html_frame_input_to_style(html_frame_styles_input: HtmlFrameStylesInput) -> HtmlFrameStyles:

    html_frame_styles_input = html_frame_styles_input if html_frame_styles_input is not None else {}

    html_frame_divider_style = set_key_resolver(instance=html_frame_styles_input, key="html_frame_divider_style", default=None)
    html_frame_divider_weight = set_key_resolver(instance=html_frame_styles_input, key="html_frame_divider_weight", default=None)
    html_frame_divider_color = set_key_resolver(instance=html_frame_styles_input, key="html_frame_divider_color", default=None)
    html_max_lines = set_key_resolver(instance=html_frame_styles_input, key="html_max_lines", default=None)
    html_multiline = set_key_resolver(instance=html_frame_styles_input, key="html_multiline", default=None)
    html_background = set_key_resolver(instance=html_frame_styles_input, key="html_background", default=None)
    html_px_multiplier = set_key_resolver(instance=html_frame_styles_input, key="html_px_multiplier", default=None)

    return HtmlFrameStyles(
        html_frame_divider_style=html_frame_divider_style,
        html_frame_divider_weight=html_frame_divider_weight,
        html_frame_divider_color=html_frame_divider_color,
        html_max_lines=html_max_lines,
        html_multiline=html_multiline,
        html_background=html_background,
        html_px_multiplier=html_px_multiplier
    )



def gen_frame_name(name: str, type: str, frame_dict: dict, ensure_unique: bool = False):
    if name is not None:
        if ensure_unique is False:
            return name
        else:
            if name in frame_dict:
                append_index = 0
                while True:
                    unique_string = f"{name}{append_index}"
                    if unique_string in frame_dict:
                        append_index += 1
                    else:
                        return unique_string

            else:
                return name
    else:
        append_index = 0
        while True:
            untitled_frame_name = f"Untited{type.capitalize()}Frame{append_index}"
            if untitled_frame_name in frame_dict:
                append_index += 1
            else:
                return untitled_frame_name



from tablate.library.ascii.chars.line_v import v_line
from tablate.library.checkers.is_last_element import is_last_element
from tablate.library.formatters.console.cell_string import cell_string_single_line, cell_string_multi_line
from tablate.library.formatters.console.row_line_color import row_colors
from tablate.library.renderers.console.frames.render_console_columns import column_console_multiline
from tablate.library.formatters.console.row_line_divider import row_line_divider
from tablate.library.formatters.console.row_outer_border import row_outer_border
from tablate.type.type_store import TableBodyFrameStore
from tablate.type.type_global import Globals


def render_console_table_frame(table_frame_store: TableBodyFrameStore, global_options: Globals) -> str:
    return_string = ""
    line_divider_string = None

    if table_frame_store.row_styles.row_line_divider != "none":
        line_divider_row_string = row_line_divider(divider=table_frame_store.row_styles.row_line_divider,
                                                   column_list_top=table_frame_store.column_list,
                                                   column_list_bottom=table_frame_store.column_list)
        line_divider_string = row_outer_border(row_string=line_divider_row_string,
                                               global_options=global_options,
                                               row_divider=table_frame_store.row_styles.row_line_divider)
    for row_index, row_item in enumerate(table_frame_store.row_list):
        if table_frame_store.frame_styles.multiline is False or table_frame_store.frame_styles.max_lines == 1:
            row_line_string = ""
            for column_index, column_item in enumerate(table_frame_store.column_list):
                row_column_item = row_colors(column_item=column_item,
                                             row_index=row_index,
                                             table_frame_store=table_frame_store)
                row_line_string += cell_string_single_line(string=row_item[column_item["key"]],
                                                           column_item=row_column_item,
                                                           column_styles=table_frame_store.column_styles,
                                                           trunc_value=table_frame_store.frame_styles.trunc_value)
                if not is_last_element(column_index, table_frame_store.column_list):
                    row_line_string += v_line[column_item["divider"]]
            return_string += row_outer_border(row_string=row_line_string, global_options=global_options)
        else:
            formatted_columns_array = []

            for column_item in table_frame_store.column_list:
                row_column_item = row_colors(column_item=column_item,
                                             row_index=row_index,
                                             table_frame_store=table_frame_store)

                column_string_array = cell_string_multi_line(string=row_item[row_column_item["key"]],
                                                             column_item=row_column_item,
                                                             column_styles=table_frame_store.column_styles,
                                                             trunc_value=table_frame_store.frame_styles.trunc_value,
                                                             max_lines=table_frame_store.frame_styles.max_lines)
                formatted_columns_array.append(column_string_array)
            return_string += column_console_multiline(formatted_columns_array=formatted_columns_array,
                                                      frame_dict=table_frame_store,
                                                      global_options=global_options)
        if not is_last_element(row_index, table_frame_store.row_list) and line_divider_string:
            return_string += line_divider_string
    # todo: if background maybe blank doesn't add first and last line spacing (maybe remove even if no background)
    if table_frame_store.row_styles.row_line_divider == "blank":
        return_string = f"{line_divider_string}{return_string}{line_divider_string}"
    return return_string



from tablate.library.formatters.html.style.attributes.divider import divider_attr
from tablate.library.formatters.html.style.attributes.margin import margin_attr
from tablate.library.formatters.html.style.attributes.space import space_attr
from tablate.type.type_global import Globals


def render_html_head(global_options: Globals) -> str:

    outer_padding = global_options.html.html_outer_styles.html_outer_padding
    frame_border_style = global_options.html.html_outer_styles.html_outer_border_style
    frame_border_weight = global_options.html.html_outer_styles.html_outer_border_weight
    frame_border_color = global_options.html.html_outer_styles.html_outer_border_color
    html_px_multiplier = global_options.html.html_outer_styles.html_px_multiplier

    html_width = global_options.html.html_outer_styles.html_outer_width
    column_baselines = global_options.html.column_baselines

    margin_left_px = outer_padding
    margin_string = f'0 {margin_left_px}px'

    border_string = divider_attr(divider_style=frame_border_style,
                                 divider_weight=frame_border_weight,
                                 divider_color=frame_border_color)

    global_options.html.styler.add_global_style_attribute("font-family", "'Roboto', sans-serif")
    global_options.html.styler.add_global_style_attribute("box-sizing", "border-box")
    global_options.html.styler.add_global_style_attribute("margin", 0)
    global_options.html.styler.add_global_style_attribute("padding", 0)
    global_options.html.styler.wrapper.add_style_attribute("cursor", "default")
    if global_options.html.html_outer_styles.html_default_colors:
        global_options.html.styler.wrapper.add_style_attribute("background-color", "transparent", " tbody tr")
        global_options.html.styler.wrapper.add_style_attribute("background-color", "rgba(255,255,255,0.95)")
        global_options.html.styler.wrapper.add_style_attribute("background-color", "transparent", " tr:nth-child(odd)")
        global_options.html.styler.wrapper.add_style_attribute("background-color", "rgba(123,123,123,0.10)", " tr:nth-child(even)")
        global_options.html.styler.wrapper.add_style_attribute("background-color", "rgba(50,123,200,0.15)", " tr:hover")
        global_options.html.styler.wrapper.add_style_attribute("background-color", "rgba(123,123,123,0.15)", " th")
        global_options.html.styler.wrapper.add_style_attribute("border", "black", " *")
        global_options.html.styler.wrapper.add_style_attribute("color", "#030303", " *")

    margin_edge = space_attr(html_spacer=margin_left_px, html_px_multiplier=html_px_multiplier)[1]

    global_options.html.styler.wrapper.add_style_attribute("width", f"calc({html_width} - {margin_edge}px)")
    global_options.html.styler.wrapper.add_style_attribute("margin", margin_attr(margin=margin_left_px,
                                                                                 html_px_multiplier=html_px_multiplier))

    global_options.html.styler.table.add_style_attribute("width", "100%")
    global_options.html.styler.table.add_style_attribute("height", "1px")
    global_options.html.styler.table.add_style_attribute("padding", "0", " tr th td")


    global_options.html.styler.table.add_style_attribute("border", border_string)

    global_options.html.styler.table.add_style_attribute("border-collapse", "collapse")

    wrapper_classes = global_options.html.styler.wrapper.generate_class_names()
    table_classes = global_options.html.styler.table.generate_class_names()

    return_string = ''

    return_string += f'<div class="{wrapper_classes}">'
    return_string += f'<table class="{table_classes}">'

    return_string += f'<colgroup>'

    previous_baseline_value = 0
    for baseline_column_width in column_baselines:
        column_width = baseline_column_width - previous_baseline_value
        return_string += f'<col style="width:{column_width}%;">'
        previous_baseline_value = baseline_column_width
    return_string += f'</colgroup>'

    return return_string




from typing import TypedDict, Dict, NotRequired, Optional

from tablate.type.primitives import TextString, FrameDivider, ColumnWidth, TextStyle, TextAlign, Background, \
    ColumnPadding, TextColor, ColumnDivider, HtmlDividerWeight, HtmlColumnDividerStyle, HtmlColumnPadding, \
    HtmlTextStyle, HtmlTextAlign, HtmlTextColor, HtmlBackground, BackgroundPadding, HtmlTextSize, MaxLines, \
    Multiline, TruncValue, HtmlFrameDivider, HtmlOuterBorder, HtmlOuterPadding, HtmlOuterWidth, HtmlDividerColor


class HtmlOuterStylesInput(TypedDict):  # HtmlOuterBase
    html_outer_border_weight: NotRequired[HtmlDividerWeight]
    html_outer_border_style: NotRequired[HtmlOuterBorder]
    html_outer_border_color: NotRequired[HtmlDividerColor]
    html_outer_padding: NotRequired[HtmlOuterPadding]
    html_outer_width: NotRequired[HtmlOuterWidth]


########################################################################################################################

class FrameStylesInput(TypedDict):  # FrameBase
    frame_divider: NotRequired[FrameDivider]
    max_lines: NotRequired[MaxLines]
    multiline: NotRequired[Multiline]
    background: NotRequired[Background]
    trunc_value: NotRequired[TruncValue]


class HtmlFrameStylesInput(TypedDict):  # HtmlFrameBase
    html_frame_divider_style: NotRequired[HtmlFrameDivider]
    html_frame_divider_weight: NotRequired[HtmlDividerWeight]
    html_frame_divider_color: NotRequired[HtmlDividerColor]
    html_max_lines: NotRequired[MaxLines]
    html_multiline: NotRequired[Multiline]
    html_background: NotRequired[HtmlBackground]


########################################################################################################################

class ColumnStylesInput(TypedDict):  # ColumnBase
    column_divider: NotRequired[ColumnDivider]
    padding: NotRequired[ColumnPadding]
    background_padding: NotRequired[BackgroundPadding]


class TextStylesInput(TypedDict):  # TextBase
    text_style: NotRequired[TextStyle]
    text_align: NotRequired[TextAlign]
    text_color: NotRequired[TextColor]


class HtmlColumnStylesInput(TypedDict):  # HtmlColumnBase
    html_column_divider_style: NotRequired[HtmlColumnDividerStyle]
    html_column_divider_weight: NotRequired[HtmlDividerWeight]
    html_column_divider_color: NotRequired[HtmlDividerColor]
    html_padding: NotRequired[HtmlColumnPadding]


class HtmlTextStylesInput(TypedDict):  # HtmlTextBase
    html_text_style: NotRequired[HtmlTextStyle]
    html_text_align: NotRequired[HtmlTextAlign]
    html_text_color: NotRequired[HtmlTextColor]
    html_text_size: NotRequired[HtmlTextSize]
    
    
class RowsStylesInput(TypedDict):  # TableRowsBase
    row_line_divider: NotRequired[FrameDivider]
    odds_background: NotRequired[Background]
    evens_background: NotRequired[Background]
    
    
class HtmlRowsStylesInput(TypedDict):  # HtmlTableRowsBase
    html_row_line_divider_weight: NotRequired[HtmlDividerWeight]
    html_row_line_divider_style: NotRequired[HtmlFrameDivider]
    html_row_line_divider_color: NotRequired[HtmlDividerColor]
    html_odds_background: NotRequired[HtmlBackground]
    html_evens_background: NotRequired[HtmlBackground]


########################################################################################################################
########################################################################################################################
########################################################################################################################

class HtmlColumnInput(TypedDict):
    padding: NotRequired[HtmlColumnPadding]
    divider_style: NotRequired[HtmlColumnDividerStyle]
    divider_weight: NotRequired[HtmlDividerWeight]
    divider_color: NotRequired[HtmlDividerColor]
    text_style: NotRequired[HtmlTextStyle]
    text_align: NotRequired[HtmlTextAlign]
    text_color: NotRequired[HtmlTextColor]
    background: NotRequired[HtmlBackground]


class BaseColumnInput(TypedDict):
    string: TextString
    width: NotRequired[ColumnWidth]
    padding: NotRequired[ColumnPadding]
    divider: NotRequired[ColumnDivider]
    text_style: NotRequired[TextStyle]
    text_align: NotRequired[TextAlign]
    text_color: NotRequired[TextColor]
    background: NotRequired[Background]
    html_styles: NotRequired[Optional[HtmlColumnInput]]


class GridColumnInput(BaseColumnInput):
    pass


class TableColumnInput(BaseColumnInput):
    key: str


TableRowsInput = Dict[str, TextString]


########################################################################################################################
########################################################################################################################
########################################################################################################################


class BaseStylesInput(TypedDict):
    frame_styles: NotRequired[FrameStylesInput]
    column_styles: NotRequired[ColumnStylesInput]
    text_styles: NotRequired[TextStylesInput]
#
#
# GridStylesInput = BaseStylesInput
#
#
TableHeaderFrameStylesInput = BaseStylesInput


class TableBodyFrameStylesInput(BaseStylesInput):
    row_styles: NotRequired[RowsStylesInput]


########################################################################################################################


class HtmlStylesInput(TypedDict):
    html_frame_styles: NotRequired[HtmlFrameStylesInput]
    html_column_styles: NotRequired[HtmlColumnStylesInput]
    html_text_styles: NotRequired[HtmlTextStylesInput]


class HtmlTextFrameStylesInput(TypedDict):
    html_frame_styles: NotRequired[HtmlFrameStylesInput]
    html_text_styles: NotRequired[HtmlTextStylesInput]


HtmlGridStylesInput = HtmlStylesInput


class HtmlTableStylesInput(HtmlStylesInput):
    html_row_styles: NotRequired[HtmlRowsStylesInput]


HtmlTableHeaderStylesInput = HtmlStylesInput

HtmlTableBodyStylesInput = HtmlTableStylesInput

########################################################################################################################
########################################################################################################################
########################################################################################################################




import functools
from copy import copy, deepcopy
from typing import Callable, Tuple
from tablate.classes.bases.TablateBase import TablateBase
from tablate.library.initializers.processors.process_frame import process_frame_args
from tablate.library.initializers.globals_init import globals_init
from tablate.library.renderers.console.render_console import render_console
from tablate.library.renderers.html.render_html import render_html

from tablate.type.primitives import OuterBorder, FrameDivider, OuterWidth, OuterPadding, Background, HtmlPxMultiplier, \
    BackgroundPadding, HtmlDefaultColors, FrameName, FrameType, GlobalArgs, FrameArgs
from tablate.type.type_global import GlobalsStore
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, \
    HtmlFrameStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import FrameDict


class TablateApiBase(TablateBase):

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 html_default_colors: HtmlDefaultColors = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        args = copy(locals())
        del args["self"]

        self._globals_store = GlobalsStore(args=args, store=globals_init(**args))
        self._css_injection = []
        self._frame_list = {}

    # def __is_ipython(self):
    #     return hasattr(__builtins__, "__IPYTHON__")

    # def list_frames(self):
    #     print_frame_list(self._frame_list)

    def apply(self, function: Callable[[FrameName, FrameType, FrameArgs, GlobalArgs], Tuple[FrameArgs, GlobalArgs]],
              raise_merge_error: bool = True):
        frame_list_copy = deepcopy(self._frame_list)
        global_options_copy = deepcopy(self._globals_store)
        for frame_copy_key, frame_copy_item in frame_list_copy.items():
            try:
                return_result = function(frame_copy_key,
                                         frame_copy_item.type,
                                         deepcopy(frame_copy_item.args),
                                         deepcopy(global_options_copy.args))
                if return_result is not None:
                    return_frame_dict, return_global_dict = return_result
                    if global_options_copy is not None:
                        self._globals_store = GlobalsStore(args=return_global_dict,
                                                           store=globals_init(**return_global_dict))
                    if return_frame_dict is not None:
                        self._frame_list[return_frame_dict["name"]] = FrameDict(name=return_frame_dict["name"],
                                                                                type=frame_copy_item.type,
                                                                                args=return_frame_dict,
                                                                                store=process_frame_args(
                                                                                    frame_args=return_global_dict,
                                                                                    frame_type=frame_copy_item.type,
                                                                                    global_options=self._globals_store.store))

            except Exception as e:
                self._frame_list = frame_list_copy
                self._globals_store = global_options_copy
                if raise_merge_error:
                    raise Exception(f"Error merging result of apply on item: {frame_copy_key}.\n\t{e}")

    def apply_style(self, selector: str, css: str, sub_selector: str = None):
        self._css_injection.append({"selector": selector, "css": css, "sub_selector": sub_selector})

    # @functools.cache
    def to_string(self) -> str:
        return render_console(frame_list=self._frame_list, global_options=self._globals_store.store)

    # @functools.cache
    def print(self) -> None:
        print(self.to_string())

    # @functools.cache
    def __repr__(self) -> str:
        return self.to_string()

    # @functools.cache
    def to_html(self) -> str:
        return render_html(frame_list=self._frame_list, global_options=self._globals_store.store,
                           css_injection=self._css_injection)

    # @functools.cache
    def _repr_html_(self) -> str:
        return self.to_html()



from typing import overload, TypeVar, Type

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.html.style.attributes.color import text_color_attr, background_color_attr
from tablate.type.defaults import column_divider_default, column_padding_default, text_style_default, \
    text_align_default, text_color_default, background_default, html_padding_default, html_divider_weight_default, \
    html_px_multiplier_default, html_divider_color_default, html_column_padding_default, html_text_style_default, \
    html_text_align_default, html_text_color_default, html_background_default, html_column_divider_style_default
from tablate.type.primitives import HtmlPxMultiplier, BackgroundPadding
from tablate.type.type_style import ColumnStyles, TextStyles, FrameStyles, HtmlFrameStyles, HtmlColumnStyles, \
    HtmlTextStyles
from tablate.type.type_input import BaseColumnInput, GridColumnInput, TableColumnInput


T = TypeVar("T", BaseColumnInput, GridColumnInput, TableColumnInput)


@overload
def column_attr(column_dict: BaseColumnInput,
                base_frame_styles: FrameStyles,
                base_column_styles: ColumnStyles,
                base_text_styles: TextStyles,
                html_frame_styles: HtmlFrameStyles,
                html_column_styles: HtmlColumnStyles,
                html_text_styles: HtmlTextStyles,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> BaseColumnInput:
    ...


@overload
def column_attr(column_dict: GridColumnInput,
                base_frame_styles: FrameStyles,
                base_column_styles: ColumnStyles,
                base_text_styles: TextStyles,
                html_frame_styles: HtmlFrameStyles,
                html_column_styles: HtmlColumnStyles,
                html_text_styles: HtmlTextStyles,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> BaseColumnInput:
    ...


@overload
def column_attr(column_dict: TableColumnInput,
                base_frame_styles: FrameStyles,
                base_column_styles: ColumnStyles,
                base_text_styles: TextStyles,
                html_frame_styles: HtmlFrameStyles,
                html_column_styles: HtmlColumnStyles,
                html_text_styles: HtmlTextStyles,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> BaseColumnInput:
    ...


def column_attr(column_dict: Type[T],
                base_frame_styles: FrameStyles,
                base_column_styles: ColumnStyles,
                base_text_styles: TextStyles,
                html_frame_styles: HtmlFrameStyles,
                html_column_styles: HtmlColumnStyles,
                html_text_styles: HtmlTextStyles,
                html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> BaseColumnInput:
    base_divider = set_attr_resolver(instance=base_column_styles, attr="column_divider", default=column_divider_default)
    base_padding = set_attr_resolver(instance=base_column_styles, attr="padding", default=column_padding_default)
    base_text_style = set_attr_resolver(instance=base_text_styles, attr="text_style", default=text_style_default)
    base_text_align = set_attr_resolver(instance=base_text_styles, attr="text_align", default=text_align_default)
    base_text_color = set_attr_resolver(instance=base_text_styles, attr="text_color", default=text_color_default)
    base_background = set_attr_resolver(instance=base_frame_styles, attr="background", default=background_default)

    html_divider_style = set_attr_resolver(instance=html_column_styles, attr="html_column_divider_style", default=html_column_divider_style_default)
    html_divider_weight = set_attr_resolver(instance=html_column_styles, attr="html_column_divider_weight", default=html_divider_weight_default)
    html_divider_color = set_attr_resolver(instance=html_column_styles, attr="html_column_divider_color", default=html_divider_color_default)
    html_padding = set_attr_resolver(instance=html_column_styles, attr="html_padding", default=html_column_padding_default)
    html_text_style = set_attr_resolver(instance=html_text_styles, attr="html_text_style", default=html_text_style_default)
    html_text_align = set_attr_resolver(instance=html_text_styles, attr="html_text_align", default=html_text_align_default)
    html_text_color = set_attr_resolver(instance=html_text_styles, attr="html_text_color", default=html_text_color_default)
    html_background = set_attr_resolver(instance=html_frame_styles, attr="html_background", default=html_background_default)

    html_styles = {}

    if "divider" in column_dict and column_dict["divider"] is not None:
        base_divider = column_dict["divider"]
        html_styles["divider_style"] = column_dict["divider"]
        html_styles["divider_weight"] = html_divider_weight * html_px_multiplier
        html_styles["divider_color"] = html_divider_color
    if "padding" in column_dict and column_dict["padding"] is not None:
        base_padding = column_dict["padding"]
        html_styles["padding"] = column_dict["padding"] * html_padding_default
    if "text_style" in column_dict and column_dict["text_style"] is not None:
        base_text_style = column_dict["text_style"]
        html_styles["text_style"] = column_dict["text_style"]
    if "text_align" in column_dict and column_dict["text_align"] is not None:
        base_text_align = column_dict["text_align"]
        html_styles["text_align"] = column_dict["text_align"]
    if "text_color" in column_dict and column_dict["text_color"] is not None:
        base_text_color = column_dict["text_color"]
        html_styles["text_color"] = text_color_attr(column_dict["text_color"])
    if "background" in column_dict and column_dict["background"] is not None:
        base_background = column_dict["background"]
        html_styles["background"] = background_color_attr(column_dict["background"])

    if "html_styles" in column_dict or ("html_styles" in column_dict and column_dict["html_styles"] is None):
        html_styles = None

    column_dict: BaseColumnInput = {
        "width": column_dict["width"],
        "string": column_dict["string"],
        "divider": base_divider,
        "padding": base_padding,
        "text_style": base_text_style,
        "text_align": base_text_align,
        "text_color": base_text_color,
        "background": base_background,
        "html_styles": html_styles if html_styles is not None else column_dict["html_styles"],
        **column_dict
    }

    return column_dict


# todo: can use TypeVar + Type[T] (hopefully) to do away with the overloads... or not...

# todo: fix this!



from typing import Union

from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.library.initializers.mappers.object.element.html_column_input_to_style import html_column_input_to_style
from tablate.library.initializers.mappers.object.element.html_frame_input_to_style import html_frame_input_to_style
from tablate.library.initializers.mappers.object.element.html_row_input_style import html_row_input_to_style
from tablate.library.initializers.mappers.object.element.html_text_input_to_style import html_text_input_to_style
from tablate.type.type_input import HtmlTableStylesInput, HtmlGridStylesInput
from tablate.type.type_store import HtmlFrameStore, HtmlTableFrameStore


def html_frame_input_to_store(html_frame_input: Union[HtmlGridStylesInput, HtmlTableStylesInput],
                              frame_type: str) -> Union[HtmlFrameStore, HtmlTableFrameStore]:
    if frame_type == "table":
        return HtmlTableFrameStore(
            html_frame_styles=html_frame_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_frame_styles",
                                 default={})),
            html_column_styles=html_column_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_column_styles",
                                 default={})),
            html_row_styles=html_row_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_row_styles",
                                 default={})),
            html_text_styles=html_text_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_text_styles",
                                 default={})))
    else:
        return HtmlFrameStore(
            html_frame_styles=html_frame_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_frame_styles",
                                 default={})),
            html_column_styles=html_column_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_column_styles",
                                 default={})),
            html_text_styles=html_text_input_to_style(
                set_key_resolver(instance=html_frame_input, key="html_text_styles",
                                 default={})))



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import FrameStylesInput
from tablate.type.type_style import FrameStyles


def base_frame_input_to_style(frame_styles_input: FrameStylesInput) -> FrameStyles:

    frame_styles_input = frame_styles_input if frame_styles_input is not None else {}

    frame_divider = set_key_resolver(instance=frame_styles_input, key="frame_divider", default=None)
    max_lines = set_key_resolver(instance=frame_styles_input, key="max_lines", default=None)
    multiline = set_key_resolver(instance=frame_styles_input, key="multiline", default=None)
    background = set_key_resolver(instance=frame_styles_input, key="background", default=None)
    trunc_value = set_key_resolver(instance=frame_styles_input, key="trunc_value", default=None)

    return FrameStyles(
        frame_divider=frame_divider,
        max_lines=max_lines,
        multiline=multiline,
        background=background,
        trunc_value=trunc_value
    )



from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import row_line_divider_default, background_default
from tablate.type.type_style import TableRowsStyles


def base_rows_mapper(rows_input: TableRowsStyles = None,
                     rows_defaults: TableRowsStyles = None) -> TableRowsStyles:
    ####################################################################################################################

    rows_input = rows_input if rows_input is not None else TableRowsStyles()
    rows_defaults = rows_defaults if rows_defaults is not None else TableRowsStyles()

    ####################################################################################################################

    row_line_divider = set_attr_resolver(instance=rows_input,
                                         attr="row_line_divider",
                                         default=set_attr_resolver(instance=rows_defaults,
                                                                   attr="row_line_divider",
                                                                   default=row_line_divider_default))
    odds_background = set_attr_resolver(instance=rows_input,
                                        attr="odds_background",
                                        default=set_attr_resolver(instance=rows_defaults,
                                                                  attr="odds_background",
                                                                  default=background_default))
    evens_background = set_attr_resolver(instance=rows_input,
                                         attr="evens_background",
                                         default=set_attr_resolver(instance=rows_defaults,
                                                                   attr="odds_background",
                                                                   default=background_default))

    rows_return = TableRowsStyles(row_line_divider=row_line_divider,
                                  odds_background=odds_background,
                                  evens_background=evens_background)

    return rows_return



from dataclasses import dataclass
from typing import Any, Type


def set_attr_resolver(instance: Any, attr: str, default: Any):
    return instance.__getattribute__(attr) if hasattr(instance, attr) and instance.__getattribute__(attr) is not None else default




from tablate.classes.bases.TablateApiBase import TablateApiBase


class TablateApiItem(TablateApiBase):

    @property
    def name(self):
        for frame_key in list(self._frame_list.items())[0]:
            return frame_key

    def rename(self, new_name: str):
        if new_name is not None:
            frame_key, frame_item = list(self._frame_list.items())[0]
            frame_item.name = new_name
            self._frame_list[new_name] = frame_item
            del self._frame_list[frame_key]


    def to_dict(self):
        frame_key, frame_item = list(self._frame_list.items())[0]
        if frame_item.type == "text":
            return {
                "text": [frame_item.args["text"]]
            }
        if frame_item.type == "grid":
            return_dict = {}
            for column_index, column_item in enumerate(frame_item.args["columns"]):
                return_dict[column_index] = [column_item["string"]]
            return return_dict
        if frame_item.type == "table":
            return_dict = {}
            for column_item in frame_item.args["columns"]:
                return_dict[column_item["key"]] = []
                for row_item in frame_item.args["rows"]:
                    return_dict[column_item["key"]].append(row_item[column_item["key"]])
            return return_dict



def divider_attr(divider_style: str, divider_weight: int, divider_color: str) -> str:
    if divider_style == "none":
        border_string = "none"
    elif divider_style == "blank":
        border_string = "none"
    elif divider_style == "thin":
        border_string = f"{divider_weight}px solid {divider_color}"
    elif divider_style == "thick":
        border_string = f"{divider_weight * 2}px solid {divider_color}"
    elif divider_style == "double":
        border_string = f"{divider_weight * 3}px double {divider_color}"
    else:
        border_string = f"{divider_weight}px solid {divider_color}"
    return border_string

# todo: seems not inheriting base styles



from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import HtmlOuterStylesInput
from tablate.type.type_style import HtmlOuterStyles


def html_outer_input_to_style(html_outer_styles_input: HtmlOuterStylesInput) -> HtmlOuterStyles:

    html_outer_styles_input = html_outer_styles_input if html_outer_styles_input is not None else {}

    html_outer_border_weight = set_key_resolver(instance=html_outer_styles_input, key="html_frame_divider_style", default=None)
    html_outer_border_style = set_key_resolver(instance=html_outer_styles_input, key="html_outer_border_style", default=None)
    html_outer_border_color = set_key_resolver(instance=html_outer_styles_input, key="html_outer_border_color", default=None)
    html_outer_padding = set_key_resolver(instance=html_outer_styles_input, key="html_outer_padding", default=None)
    html_outer_width = set_key_resolver(instance=html_outer_styles_input, key="html_outer_width", default=None)
    html_px_multiplier = set_key_resolver(instance=html_outer_styles_input, key="html_px_multiplier", default=None)
    html_default_colors = set_key_resolver(instance=html_outer_styles_input, key="html_default_colors", default=None)

    return HtmlOuterStyles(
        html_outer_border_weight=html_outer_border_weight,
        html_outer_border_style=html_outer_border_style,
        html_outer_border_color=html_outer_border_color,
        html_outer_padding=html_outer_padding,
        html_outer_width=html_outer_width,
        html_px_multiplier=html_px_multiplier,
        html_default_colors=html_default_colors
    )


from typing import Union

from tablate.type.type_store import FrameDict, FrameDictList


def select_frame(frame_list: FrameDictList, selector: Union[int, str]) -> FrameDict:
    for frame_index, (frame_key, frame_item) in enumerate(frame_list.items()):
        if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
            return frame_item



from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import background_default, max_lines_default, \
    multiline_default, trunc_value_default, frame_divider_default
from tablate.type.type_style import FrameStyles
from tablate.type.type_store import FrameStore


def base_frame_mapper(frame_input: FrameStyles = None,
                      frame_defaults: FrameStyles = None) -> FrameStyles:
    ####################################################################################################################

    frame_input = frame_input if frame_input is not None else FrameStyles()
    frame_defaults = frame_defaults if frame_defaults is not None else FrameStyles()

    ####################################################################################################################

    frame_return = FrameStyles(frame_divider=set_attr_resolver(instance=frame_input,
                                                               attr="frame_divider",
                                                               default=set_attr_resolver(instance=frame_defaults,
                                                                                         attr="frame_divider",
                                                                                         default=frame_divider_default)),
                               max_lines=set_attr_resolver(instance=frame_input,
                                                           attr="max_lines",
                                                           default=set_attr_resolver(instance=frame_defaults,
                                                                                     attr="max_lines",
                                                                                     default=max_lines_default)),
                               multiline=set_attr_resolver(instance=frame_input,
                                                           attr="multiline",
                                                           default=set_attr_resolver(instance=frame_defaults,
                                                                                     attr="multiline",
                                                                                     default=multiline_default)),
                               background=set_attr_resolver(instance=frame_input,
                                                            attr="background",
                                                            default=set_attr_resolver(instance=frame_defaults,
                                                                                      attr="background",
                                                                                      default=background_default)),
                               trunc_value=set_attr_resolver(instance=frame_input,
                                                             attr="trunc_value",
                                                             default=set_attr_resolver(instance=frame_defaults,
                                                                                       attr="trunc_value",
                                                                                       default=trunc_value_default)))

    return frame_return



from typing import List

from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.type.primitives import HtmlCellElement


def html_column_head_formatter(column_styler: ElementStyler,
                               column_index: int,
                               colspans: List[int],
                               column_element: HtmlCellElement = 'td'):

    return_html = ''

    column_classnames = column_styler.generate_class_names()

    return_html += f'<{column_element} colspan="{colspans[column_index]}" class="{column_classnames}">'

    return return_html


def html_column_foot_formatter(column_element: HtmlCellElement = 'td'):
    return f'</{column_element}>'



from copy import deepcopy
from typing import Union
from tablate.classes.bases.TablateItem import TablateItem
from tablate.classes.helpers.select_frame import select_frame

from tablate.type.type_store import FrameDictList


def get_frame(frame_list: FrameDictList, selector: Union[int, str], global_options: dict = None):
    frame_dict = deepcopy(select_frame(frame_list, selector))
    if global_options is not None:
        return TablateItem(frame_item=frame_dict, **global_options)
    else:
        return TablateItem(frame_item=frame_dict)








    # selector_index = 0
    # for frame_index, frame_item in enumerate(frame_list):
    #     if frame_item.type == "table_header":
    #         continue
    #     if (type(selector) == int and selector == selector_index) or (
    #             type(selector) == str and selector == frame_item.name):
    #         if frame_item.type == "text":
    #             return Text(text=frame_item.column_list[0]["string"],
    #                         name=frame_item.name,
    #                         text_style=frame_item.text_styles.text_style,
    #                         text_align=frame_item.text_styles.text_align,
    #                         text_color=frame_item.text_styles.text_color,
    #                         frame_divider=frame_item.frame_styles.frame_divider,
    #                         background=frame_item.frame_styles.background,
    #                         background_padding=frame_item.column_list[0]["background_padding"],
    #                         multiline=frame_item.frame_styles.multiline,
    #                         max_lines=frame_item.frame_styles.max_lines,
    #
    #                         html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
    #                         html_styles=HtmlTextFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
    #                                                              html_text_styles=frame_item.html_text_styles),
    #                         # TablateApi arge
    #                         outer_border=globals.console.outer_styles.outer_border,
    #                         outer_padding=globals.console.outer_styles.outer_padding,
    #                         outer_width=globals.console.outer_styles.outer_width,
    #                         html_outer_styles=globals.html.html_outer_styles)
    #         if frame_item.type == "grid":
    #             return Grid(columns=frame_item.column_list,
    #                         name=frame_item.name,
    #                         frame_divider=frame_item.frame_styles.frame_divider,
    #                         background=frame_item.frame_styles.background,
    #                         background_padding=frame_item.column_styles.background_padding,
    #                         multiline=frame_item.frame_styles.multiline,
    #                         max_lines=frame_item.frame_styles.max_lines,
    #                         column_divider=frame_item.column_styles.divider,
    #                         column_padding=frame_item.column_styles.padding,
    #                         text_style=frame_item.text_styles.text_style,
    #                         text_align=frame_item.text_styles.text_align,
    #                         text_color=frame_item.text_styles.text_color,
    #                         html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
    #                         html_styles=HtmlGridFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
    #                                                              html_column_styles=frame_item.html_column_styles,
    #                                                              html_text_styles=frame_item.html_text_styles),
    #                         # TablateApi args
    #                         outer_border=globals.console.outer_styles.outer_border,
    #                         outer_padding=globals.console.outer_styles.outer_padding,
    #                         outer_width=globals.console.outer_styles.outer_width,
    #                         html_outer_styles=globals.html.html_outer_styles)
    #         elif frame_item.type == "table_body":
    #             header_base_divider = None
    #             header_styles = None
    #             html_header_styles = None
    #             if frame_item.hide_header is False:
    #                 header_frame = frame_list[frame_index - 1]
    #                 header_base_divider = header_frame.frame_styles.frame_divider
    #                 header_styles = TableHeaderFrameStylesInput(frame_styles=header_frame.frame_styles,
    #                                                             column_styles=header_frame.column_styles,
    #                                                             text_styles=header_frame.text_styles)
    #                 html_header_styles = HtmlTableHeaderFrameStylesInput(
    #                     html_frame_styles=header_frame.html_frame_styles,
    #                     html_column_styles=header_frame.html_column_styles,
    #                     html_text_styles=header_frame.html_text_styles)
    #             return Table(columns=frame_item.column_list,
    #                          rows=frame_item.row_list,
    #                          name=frame_item.name,
    #                          frame_divider=frame_item.frame_styles.frame_divider,
    #                          background=frame_item.frame_styles.background,
    #                          background_padding=frame_item.column_styles.background_padding,
    #                          multiline=frame_item.frame_styles.multiline,
    #                          max_lines=frame_item.frame_styles.max_lines,
    #                          column_divider=frame_item.column_styles.divider,
    #                          column_padding=frame_item.column_styles.padding,
    #                          text_style=frame_item.text_styles.text_style,
    #                          text_align=frame_item.text_styles.text_align,
    #                          text_color=frame_item.text_styles.text_color,
    #                          header_base_divider=header_base_divider,
    #                          hide_header=frame_item.hide_header,
    #                          row_line_divider=frame_item.row_styles.row_line_divider,
    #                          odd_row_background=frame_item.row_styles.odds_background,
    #                          even_row_background=frame_item.row_styles.evens_background,
    #                          html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
    #                          header_styles=header_styles,
    #                          body_styles=TableBodyFrameStylesInput(frame_styles=frame_item.frame_styles,
    #                                                                column_styles=frame_item.column_styles,
    #                                                                text_styles=frame_item.text_styles,
    #                                                                row_styles=frame_item.row_styles),
    #                          html_header_styles=html_header_styles,
    #                          html_body_styles=HtmlTableBodyFrameStylesInput(
    #                              html_frame_styles=frame_item.html_frame_styles,
    #                              html_column_styles=frame_item.html_column_styles,
    #                              html_text_styles=frame_item.html_text_styles,
    #                              html_row_styles=frame_item.html_row_styles),
    #                          html_styles=HtmlTableFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
    #                                                                html_column_styles=frame_item.html_column_styles,
    #                                                                html_text_styles=frame_item.html_text_styles,
    #                                                                html_row_styles=frame_item.html_row_styles),
    #                          # TablateApi args
    #                          outer_border=globals.console.outer_styles.outer_border,
    #                          outer_padding=globals.console.outer_styles.outer_padding,
    #                          outer_width=globals.console.outer_styles.outer_width,
    #                          html_outer_styles=globals.html.html_outer_styles)
    #     selector_index += 1



from copy import deepcopy

from tablate.type.type_input import BaseColumnInput
from tablate.type.type_store import TableBodyFrameStore


def row_colors(column_item: BaseColumnInput, row_index: int, table_frame_store: TableBodyFrameStore) -> BaseColumnInput:
    row_column_item = deepcopy(column_item)
    if row_index % 2 == 1 and table_frame_store.row_styles.evens_background is not None:
        row_column_item["background"] = table_frame_store.row_styles.evens_background
    if row_index % 2 == 0 and table_frame_store.row_styles.odds_background is not None:
        row_column_item["background"] = table_frame_store.row_styles.odds_background
    return row_column_item



from typing import Union

from tablate.library.calcs.calc_column_widths import calc_column_widths
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.initializers.mappers.element.style_mapper import style_mapper
from tablate.library.initializers.mappers.attribute.column_attr import column_attr
from tablate.type.defaults import column_padding_default, background_padding_default, frame_name_default
from tablate.type.type_style import HtmlColumnStyles, HtmlFrameStyles, HtmlTextStyles, FrameStyles, TextStyles
from tablate.type.type_global import Globals
from tablate.type.type_input import HtmlTextFrameStylesInput, BaseColumnInput
from tablate.type.primitives import FrameDivider, TextAlign, TextStyle, TextColor, Background, Multiline, MaxLines, \
    ColumnPadding, HtmlPxMultiplier, BackgroundPadding, FrameName
from tablate.type.type_store import GridFrameStore, BaseFrameStore, HtmlFrameStore


def text_init(text: Union[str, int, float],
              name: FrameName,
              text_style: TextStyle,
              text_align: TextAlign,
              text_color: TextColor,

              frame_divider: FrameDivider,
              frame_padding: ColumnPadding,
              background: Background,
              background_padding: BackgroundPadding,
              multiline: Multiline,
              max_lines: MaxLines,

              html_px_multiplier: HtmlPxMultiplier,
              html_styles: HtmlTextFrameStylesInput,

              global_options: Globals) -> GridFrameStore:


    if html_styles is not None:
        pass

    background_padding = background_padding if background_padding is not None else set_attr_resolver(
        instance=global_options.console.outer_styles,
        attr="background_padding",
        default=background_padding_default)

    name = name if name is not None else frame_name_default

    html_frame_styles = html_styles.html_frame_styles if hasattr(html_styles,
                                                                 'html_frame_styles') else HtmlFrameStyles()
    html_column_styles = HtmlColumnStyles()
    html_text_styles = html_styles.html_text_styles if hasattr(html_styles, 'html_text_styles') else HtmlTextStyles()

    text_frame_styles = style_mapper(base_input=BaseFrameStore(frame_styles=FrameStyles(frame_divider=frame_divider,
                                                                                        max_lines=max_lines,
                                                                                        multiline=multiline,
                                                                                        background=background),
                                                               text_styles=TextStyles(text_style=text_style,
                                                                                      text_align=text_align,
                                                                                      text_color=text_color)),
                                     html_input=HtmlFrameStore(html_frame_styles=html_frame_styles,
                                                               html_column_styles=html_column_styles,
                                                               html_text_styles=html_text_styles),
                                     base_defaults=BaseFrameStore(),
                                     html_defaults=HtmlFrameStore(),
                                     html_px_multiplier=html_px_multiplier,
                                     global_options=global_options)

    # text_column: BaseColumnInput = {
    #     "string": text,
    #     "padding": frame_padding if frame_padding else global_options.console.column_styles.padding if global_options.console.column_styles.padding is not None else column_padding_default,
    # }

    columns = calc_column_widths(columns=[text_column], global_options=global_options)
    # todo: pesky warning
    column = column_attr(column_dict=columns[0],
                         base_frame_styles=text_frame_styles.frame_styles,
                         base_column_styles=text_frame_styles.column_styles,
                         base_text_styles=text_frame_styles.text_styles,
                         background_padding=background_padding)

    text_frame_store = GridFrameStore(type="text",
                                      name=name,
                                      column_list=[column],
                                      frame_styles=text_frame_styles.frame_styles,
                                      column_styles=text_frame_styles.column_styles,
                                      text_styles=text_frame_styles.text_styles,
                                      html_frame_styles=text_frame_styles.html_frame_styles,
                                      html_column_styles=text_frame_styles.html_column_styles,
                                      html_text_styles=text_frame_styles.html_text_styles)

    return text_frame_store



from tablate.library.renderers.console.render_console_foot import render_console_foot
from tablate.library.renderers.console.render_console_frames import render_console_frames
from tablate.library.renderers.console.render_console_head import render_console_head
from tablate.type.type_store import FrameDictList
from tablate.type.type_global import Globals


def render_console(frame_list: FrameDictList, global_options: Globals) -> str:
    processed_frame_list = []
    for _, frame_item in frame_list.items():
        if frame_item.type == "table":
            if frame_item.args["hide_header"] is not True:
                processed_frame_list.append(frame_item.store[0])
            processed_frame_list.append(frame_item.store[1])
        else:
            processed_frame_list.append(frame_item.store)
    return_string = ""
    if len(processed_frame_list) > 0:
        return_string += render_console_head(frame_list=processed_frame_list, global_options=global_options)
        return_string += render_console_frames(frame_list=processed_frame_list, global_options=global_options)
        return_string += render_console_foot(frame_list=processed_frame_list, global_options=global_options)
    return return_string



from typing import Callable, List, Union, Optional, Tuple

from tablate.classes.options.html.style.utilities.style_types import SelectorDictUnion, StyleItem


class AddStyleMixin:

    # todo: possibly in the future create specific methods for each style type... (ie: text-align / padding / etc)

    _selector_dict: SelectorDictUnion
    _create_style: Callable[[SelectorDictUnion, StyleItem], None]

    def add_style_attribute(self, attribute: str, value: Union[str, int], sub_selector: str = None) -> None:
        self._create_style(self._selector_dict, StyleItem(style=f"{attribute}:{value}", pseudo=sub_selector))





from dataclasses import dataclass, field
from typing import List, Optional

from tablate.classes.options.html.style.CssStyler import CssStyler
from tablate.type.primitives import CssStyleBlock, PxInteger
from tablate.type.type_style import FrameStyles, ColumnStyles, TextStyles, HtmlFrameStyles, \
    HtmlColumnStyles, HtmlTextStyles, OuterStyles, HtmlOuterStyles


@dataclass
class ConsoleGlobals:
    outer_styles: OuterStyles = field(default_factory=OuterStyles)
    frame_styles: FrameStyles = field(default_factory=FrameStyles)
    column_styles: ColumnStyles = field(default_factory=ColumnStyles)
    text_styles: TextStyles = field(default_factory=TextStyles)


@dataclass
class HtmlGlobals:
    html_outer_styles: HtmlOuterStyles = field(default_factory=HtmlOuterStyles)
    html_frame_styles: HtmlFrameStyles = field(default_factory=HtmlFrameStyles)
    html_column_styles: HtmlColumnStyles = field(default_factory=HtmlColumnStyles)
    html_text_styles: HtmlTextStyles = field(default_factory=HtmlTextStyles)
    css_injection: CssStyleBlock = ""
    styler: Optional[CssStyler] = field(default_factory=CssStyler)
    column_baselines: List[PxInteger] = field(default_factory=list[6])


@dataclass
class Globals:
    console: ConsoleGlobals = field(default_factory=ConsoleGlobals)
    html: HtmlGlobals = field(default_factory=HtmlGlobals)

@dataclass
class GlobalsStore:
    args: dict
    store: Globals


from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import HtmlColumnStylesInput
from tablate.type.type_style import HtmlColumnStyles


def html_column_input_to_style(html_column_styles_input: HtmlColumnStylesInput) -> HtmlColumnStyles:

    html_column_styles_input = html_column_styles_input if html_column_styles_input is not None else {}

    html_column_divider_style = set_key_resolver(instance=html_column_styles_input, key="html_column_divider_style", default=None)
    html_column_divider_weight = set_key_resolver(instance=html_column_styles_input, key="html_column_divider_weight", default=None)
    html_column_divider_color = set_key_resolver(instance=html_column_styles_input, key="html_column_divider_color", default=None)
    html_padding = set_key_resolver(instance=html_column_styles_input, key="html_padding", default=None)

    return HtmlColumnStyles(
        html_column_divider_style=html_column_divider_style,
        html_column_divider_weight=html_column_divider_weight,
        html_column_divider_color=html_column_divider_color,
        html_padding=html_padding
    )



from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.html.style.attributes.color import text_color_attr
from tablate.type.defaults import text_style_default, text_align_default, text_color_default, html_text_size_default, \
    html_px_multiplier_default
from tablate.type.primitives import HtmlPxMultiplier
from tablate.type.type_style import HtmlTextStyles, TextStyles


def html_text_mapper(html_text_input: HtmlTextStyles = None,
                     html_text_defaults: HtmlTextStyles = None,
                     base_text_defaults: TextStyles = None,
                     html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> HtmlTextStyles:
    if html_text_input is None:
        html_text_input = HtmlTextStyles()

    html_text_style = set_attr_resolver(
        instance=html_text_input,
        attr="html_text_style",
        default=set_attr_resolver(
            instance=html_text_defaults,
            attr="html_text_style",
            default=set_attr_resolver(
                instance=base_text_defaults,
                attr="text_style",
                default=text_style_default)))
    html_text_align = set_attr_resolver(
        instance=html_text_input,
        attr="html_text_align",
        default=set_attr_resolver(
            instance=html_text_defaults,
            attr="html_text_align",
            default=set_attr_resolver(
                instance=base_text_defaults,
                attr="text_align",
                default=text_align_default)))
    html_text_color = set_attr_resolver(
        instance=html_text_input,
        attr="html_text_color",
        default=set_attr_resolver(
            instance=html_text_defaults,
            attr="html_text_color",
            default=text_color_attr(set_attr_resolver(
                instance=base_text_defaults,
                attr="text_color",
                default=text_color_default))))
    html_text_size = set_attr_resolver(
        instance=html_text_input,
        attr="html_text_size",
        default=set_attr_resolver(
            instance=html_text_defaults,
            attr="html_text_size",
            default=html_text_size_default))

    text_return = HtmlTextStyles(html_text_style=html_text_style,
                                 html_text_align=html_text_align,
                                 html_text_color=html_text_color,
                                 html_text_size=html_text_size)

    return text_return



from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import column_divider_default, column_padding_default
from tablate.type.type_style import ColumnStyles


def base_column_mapper(columns_input: ColumnStyles = None,
                       column_defaults: ColumnStyles = None) -> ColumnStyles:
    ####################################################################################################################

    columns_input = columns_input if columns_input is not None else ColumnStyles()
    column_defaults = column_defaults if column_defaults is not None else ColumnStyles()

    ####################################################################################################################

    columns_return = ColumnStyles(column_divider=set_attr_resolver(instance=columns_input,
                                                                   attr="divider",
                                                                   default=set_attr_resolver(instance=column_defaults,
                                                                                      attr="divider",
                                                                                      default=column_divider_default)),
                                  padding=set_attr_resolver(instance=columns_input,
                                                            attr="padding",
                                                            default=set_attr_resolver(instance=column_defaults,
                                                                                      attr="padding",
                                                                                      default=column_padding_default)),
                                  background_padding=set_attr_resolver(instance=columns_input,
                                                                       attr="background_padding",
                                                                       default=set_attr_resolver(
                                                                           instance=column_defaults,
                                                                           attr="background_padding",
                                                                           default=column_padding_default))
                                  )

    return columns_return



from typing import List

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.initializers.processors.process_frame import process_frame
from tablate.type.primitives import FrameDivider, ColumnDivider, OuterBorder, \
    FrameName, Multiline, MaxLines, Background, BackgroundPadding, HideHeader, ColumnPadding, TextAlign, TextStyle, \
    TextColor, HtmlPxMultiplier, OuterPadding, OuterWidth, HtmlDefaultColors
from tablate.type.type_input import TableColumnInput, TableRowsInput, TableBodyFrameStylesInput, \
    TableHeaderFrameStylesInput, HtmlTableStylesInput, HtmlTableHeaderStylesInput, \
    HtmlTableBodyStylesInput, HtmlOuterStylesInput


class Table(TablateApiItem):

    def __init__(self,
                 # TablateTable args
                 columns: List[TableColumnInput],
                 rows: List[TableRowsInput],

                 name: FrameName = None,

                 frame_divider: FrameDivider = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 multiline_header: Multiline = None,
                 max_lines_header: MaxLines = None,
                 hide_header: HideHeader = None,

                 column_divider: ColumnDivider = None,
                 column_padding: ColumnPadding = None,
                 header_base_divider: FrameDivider = None,

                 row_line_divider: FrameDivider = None,
                 odd_row_background: Background = None,
                 even_row_background: Background = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,

                 header_styles: TableHeaderFrameStylesInput = None,
                 body_styles: TableBodyFrameStylesInput = None,

                 html_default_colors: HtmlDefaultColors = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTableStylesInput = None,

                 html_header_styles: HtmlTableHeaderStylesInput = None,
                 html_body_styles: HtmlTableBodyStylesInput = None,

                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                outer_border=outer_border,
                                outer_padding=outer_padding,
                                frame_divider=frame_divider,
                                outer_width=outer_width,
                                html_default_colors=html_default_colors,
                                html_outer_styles=html_outer_styles)

        frame_dict = process_frame(frame_name=name,
                                   frame_type="table",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict



from copy import deepcopy
from typing import List

from tablate.classes.bases.TablateSet import TablateSet
from tablate.classes.classes import TablateUnion
from tablate.type.primitives import FrameDivider, OuterBorder, OuterPadding, OuterWidth, Background, BackgroundPadding, \
    HtmlPxMultiplier
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import FrameDictList


def concat(frame_list: List[TablateUnion],
           outer_border: OuterBorder = None,
           outer_padding: OuterPadding = None,
           outer_width: OuterWidth = None,

           frame_divider: FrameDivider = None,
           background: Background = None,
           background_padding: BackgroundPadding = None,

           html_px_multiplier: HtmlPxMultiplier = None,
           html_outer_styles: HtmlOuterStylesInput = None,

           column_styles: ColumnStylesInput = None,
           text_styles: TextStylesInput = None,

           html_frame_styles: HtmlFrameStylesInput = None,

           html_column_styles: HtmlColumnStylesInput = None,
           html_text_styles: HtmlTextStylesInput = None):

    input_dict_list: List[FrameDictList] = []

    for frame_set in frame_list:
        input_dict_list.append(deepcopy(frame_set._frame_list))

    args = deepcopy(locals())

    args["frame_list"] = input_dict_list
    del args["input_dict_list"]
    del args["frame_set"]

    return TablateSet(**args)

# todo: this needs to have process frame applied (probably in TabletSet)
